"""Classes for concatenating and batch-processing multiple recordings via
Suite2p and SIMA"""

import os
import pickle
import sima

from . import Suite2pImagingDataset
from ...signal import SignalFile


class BatchMixin:

    """Mixin class that modifies the functionality of sima ImagingDatasets.
    This facilitates the construction of multi-sequence datasets from existing
    sima datasets, for batch processing. The directories of the original child
    datasets are retained, so that results can be later applied to those
    datasets. Datasets derived from BatchMixin can be initialized in several
    ways, depending on the passed parameters.

    Parameters
    ----------
    sequences : list of Sequence objects, optional
        Sequences to include in the batch dataset. If passed, you must also
        pass 'child_dirs' (see below).
    child_dirs : list of str, optional
        A list of sima directories. If 'sequences' is passed, this is a list of
        the same length; a sima ImagingDataset will be created for each
        sequence and saved in these directories. If 'child_dirs' is passed
        alone, the batch dataset will be created from existing sima datasets
        at these locations (if they exist).
    ds_list : list of ImagingDataset objects, optional
        The batch dataset is created by combining sequences from the passed
        ImagingDataset objects.
    savedir : str, optional
        Where to save the batch dataset
    channel_names : list of str, optional
        If not passed, the channel names are taken from the first child
        dataset in the batch.

    Attributes
    ----------
    child_dirs : list of str
        A list of sima directories, corresponding to each of the child
        datasets
    """

    def __init__(self, sequences=None, savedir=None, child_dirs=None,
                 ds_list=None, channel_names=None):

        # TODO Is there a more pythonic way of doing this?
        try:
            self.base_class = self.__class__.__bases__[1]
        except Exception:
            self.base_class = sima.ImagingDataset

        if (not sequences) and (not child_dirs) and (not ds_list):
            # Load an existing batch dataset
            super().__init__(None, savedir)

            with open(os.path.join(savedir, 'dataset.pkl'), 'rb') as f:
                data = pickle.load(f)

            # Add child directory information as an attribute
            try:
                self.child_dirs = data.pop('child_dirs')
            except KeyError:
                raise Exception(f'{savedir} is not a batched dataset')
            return

        elif sequences:
            # Create a new batch dataset from sequences, and create child sima
            # folders
            if (child_dirs is None) or (len(child_dirs) != len(sequences)):
                raise Exception("A child directory must be specified for "
                                + "every sequence")
            elif ds_list:
                raise ValueError("Either 'sequences' or 'ds_list' must be "
                                 + "specified, but not both")
            else:
                # create child sima directories
                if any([os.path.exists(child_dir)
                        for child_dir in child_dirs]):
                    raise Exception("At least one child directory already "
                                    + "exists")
                ds_list = [self.base_class(sequence, child_dir,
                                           channel_names=channel_names)
                           for sequence, child_dir in
                           zip(*[sequences, child_dirs])]

        elif child_dirs:
            # Create a new batch dataset from existing sima directories
            if ds_list:
                raise Exception('TODO')
            else:
                ds_list = [self.base_class.load(child_dir)
                           for child_dir in child_dirs]

        if ds_list:
            # Create batch dataset
            if any([len(ds.sequences) > 1 for ds in ds_list]):
                raise Exception("Each child dataset can only have one "
                                + "sequence")
            elif len(set([ds.frame_shape for ds in ds_list])) > 1:
                raise Exception("All datasets must have compatible frame "
                                + "shapes")
            else:
                if channel_names is None:
                    channel_names = ds_list[0].channel_names
                sequences = [ds.sequences[0] for ds in ds_list]
                self.child_dirs = [ds.savedir for ds in ds_list]
                super().__init__(sequences, savedir,
                                 channel_names=channel_names)
        else:
            raise TypeError("New batch datasets must be initialized from "
                            + "'sequences'/'child_dirs', 'child_dirs', or "
                            + "'ds_list'.")

    def _todict(self):
        return {**super()._todict(), **{'child_dirs': self.child_dirs}}

    @property
    def children(self):
        child_datasets = []
        for path in self.child_dirs:
            child_datasets.append(self.base_class.load(path))
        return child_datasets


class BatchSuite2pImagingDataset(BatchMixin, Suite2pImagingDataset):

    """Class for running batch processing of recordings via Suite2p. Datasets
    are concatenated and jointly motion corrected, if desired. ROI detection
    and signal extraction are performed on the concatenated dataset. Following
    curation of ROIs in the Suite2p GUI, the class provides a method to
    split the signals and farm them back to the original child datasets.

    Parameters
    ----------
    sequences : list of Sequence objects, optional
        Sequences to include in the batch dataset. If passed, you must also
        pass 'child_dirs' (see below).
    child_dirs : list of str, optional
        A list of sima directories. If 'sequences' is passed, this is a list of
        the same length; a sima ImagingDataset will be created for each
        sequence and saved in these directories. If 'child_dirs' is passed
        alone, the batch dataset will be created from existing sima datasets
        at these locations (if they exist).
    ds_list : list of ImagingDataset objects, optional
        The batch dataset is created by combining sequences from the passed
        ImagingDataset objects.
    savedir : str, optional
        Where to save the batch dataset
    channel_names : list of str, optional
        If not passed, the channel names are taken from the first child
        dataset in the batch.

    Attributes
    ----------
    child_dirs : list of str
        A list of sima directories, corresponding to each of the child
        datasets
    """

    def import_results_to_children(self, label='suite2p', channel='Ch2',
                                   overwrite=False):
        """Split concatenated Suite2p results amongst the child datasets
        associated with this batched dataset.

        Parameters
        ----------
        label : str, optional
            ROI label to store the signals and ROI masks under. Defaults to
            'suite2p'.
        channel : {str or int}, optional
            Name or index of dynamic channel in the imaging dataset. Defaults
            to 'Ch2'.
        overwrite : bool, optional
            If key '/{channel}/{label}' already exists in the dataset signals
            file, overwrite with new import. Note this will delete any
            additional signal types in the store in addition to the 'raw' and
            'npil' (e.g. 'dfof'), to prevent mixtures of signals derived from
            different imports. Defaults to False, which will raise an error if
            the desired key already exists.
        """

        # Maybe functionalize repeated elements from the standard import
        # method

        # if the concatenated results have not been imported, do so
        if not os.path.exists(os.path.join(self.savedir, "signals.h5")):
            print("No parent signals found. Importing parent dataset first:")
            self.import_results(label=label, channel=channel,
                                overwrite=overwrite)

        # find all signals records for matching the desired channel/label
        with SignalFile(os.path.join(self.savedir, "signals.h5")) \
                as parent_signal_file:

            base_key = f"/Ch2/{label}"
            keys_to_split = [key for key in parent_signal_file.keys()
                             if base_key in key]
            data_to_split = {key: parent_signal_file[key]
                             for key in keys_to_split}

            if len(keys_to_split):
                print(f"Signal keys to split among children: {keys_to_split}")
            else:
                raise KeyError(f"No parent signals found under {base_key}")

        # slice concatenated signals and save under each child sima dataset
        frame_count = 0
        for child in self.children:
            child_slice = slice(frame_count, frame_count + child.num_frames)
            print(f"\n{child.savedir} <--> {child_slice}")
            frame_count += child.num_frames

            with SignalFile(os.path.join(child.savedir, "signals.h5")) \
                    as child_signal_file:

                for key, df in data_to_split.items():
                    if key in child_signal_file:
                        assert overwrite, f"Signals already exist for {key}." \
                            + " Set overwrite=True to re-import.\nCurrent " \
                            + f"file structure:\n {child_signal_file.info()}"

                        print(f"Deleting previous import at {key}")
                        child_signal_file.remove(key)

                    # add signals to child dataset
                    child_signal_file.put(
                        key,
                        df.iloc[:, child_slice].T.reset_index(drop=True).T)

                print(child_signal_file.info())

            # add ROIs to child dataset (TODO: another overwrite check?)
            child.add_ROIs(self.ROIs[label], label)

    def apply_mc_to_children(self, overwrite=True):
        """doc string"""

        frame_count = 0
        for child in self.children:
            child_slice = slice(frame_count, frame_count + child.num_frames)
            print(f"\n Applying mc to {child.savedir} <--> {child_slice}")

            frame_count += child.num_frames
            child_ops = self.ops_list
            for plane in range(len(child_ops)):
                child_ops[plane]['xoff'] = \
                    child_ops[plane]['xoff'][child_slice]
                child_ops[plane]['yoff'] = \
                    child_ops[plane]['yoff'][child_slice]

            child.apply_mc_results(overwrite=overwrite, ops_list=child_ops)



import os

import sima
import pandas as pd 
import numpy as np
from scipy.ndimage import center_of_mass

from lab3.signal import SignalFile
from lab3.extraction.cnmf import AROIs

def import_to_signals_file(ds, label='cnmf', channel='Ch2',
                           overwrite=False):
    """Import CNMF signal results using the new-style signal formats
    (dataframes and h5 stores).

    Parameters
    ----------
    ds : CNMFDataset
        Instance of a CNMFDataset, which has been previously extracted.
    label : str, optional
        ROI label to store the signals and ROI masks under. Defaults to
        'cnmf'.
    channel : {str or int}, optional
        Name or index of dynamic channel in the imaging dataset. Defaults to
        'Ch2'.
    overwrite : bool, optional
        If key '/{channel}/{label}' already exists in the dataset signals file,
        overwrite with new import. Note this will delete any additional signal
        types in the store in addition to the 'raw' and 'npil' (e.g. 'dfof'),
        to prevent mixtures of signals derived from different imports. Defaults
        to False, which will raise an error if the desired key already exists.
    """

    # check if key already exists in signals file, and deal with overwrites
    with SignalFile(os.path.join(ds.savedir, "signals.h5")) as signal_file:
        base_key = f"/{channel}/{label}"
        if base_key in signal_file:
            assert overwrite, f"Signals already exist for {base_key}. " \
                + "Set overwrite=True to re-import.\nCurrent file structure:" \
                + f" \n {signal_file.info()}"

            print(f"Deleting previous import at {base_key}")
            signal_file.remove(base_key)

        # create a sima ROI list and save to the underlying imaging dataset
        print("Making rois...")
        rois = _make_cell_rois(ds)
        print(f"Made {len(rois)} rois")
        ds.add_ROIs(rois, label=label)

        background_rois = _make_background_rois(ds)
        print(f"Made {len(background_rois)} background components")
        ds.add_ROIs(background_rois, label='background')

        # create signal dataframes
        # TODO replace this with proper signal classes for each type,
        # and store metadata
        idx = pd.Index([roi.label for roi in rois], name='roi_label')
        bkd_idx = pd.Index([roi.label for roi in background_rois], 
                            name='roi_label')

        raw = pd.DataFrame(ds.cnm.estimates.C, index=idx) 
        dfof = pd.DataFrame(ds.cnm.estimates.F_dff, index=idx)
        background = pd.DataFrame(ds.cnm.estimates.f, index=bkd_idx)

        if ds.cnm.params.temporal['p'] > 0:
            spikes = pd.DataFrame(ds.cnm.estimates.S, index=idx)

        # store signals
        print("Storing...")
        signal_file.put(f"{base_key}/raw", raw)
        signal_file.put(f"{base_key}/dfof", dfof)        
        signal_file.put(f"{base_key}/background", background)
        if ds.cnm.params.temporal['p'] > 0:
            signal_file.put(f"{base_key}/spikes", spikes)

        print(signal_file.info())

def name_rois(arois):
    names = []
    for r in arois:
        com = center_of_mass(r.tensor)
        hashes = [f"{hash(coord)%10**4:04}" for coord in com]
        name = '-'.join(hashes)
        names.append(name)
    return names

def _make_cell_rois(ds):
    arois = AROIs.AROIs(matrix=ds.cnm.estimates.A, dims=ds.dims)
    roi_labels = name_rois(arois)
    arois = AROIs.AROIs(matrix=arois.matrix, labels=roi_labels, dims=ds.dims)
    rois = arois.roilist

    return rois

def _make_background_rois(ds):
    num_background = ds.cnm.params.temporal['nb']
    background_labels = [f"BACKGROUND{i}" for i in range(num_background)]
    background_arois = AROIs.AROIs(matrix=ds.cnm.estimates.b, 
                                    dims=ds.dims, labels=background_labels)
    background_rois = background_arois.roilist

    return background_rois



"""A collection of classes for representing groups of experiments."""

from functools import reduce
from collections import defaultdict

from lab3.core import Group
from lab3.experiment.base import (fetch_trials, _BaseExperiment,
                                  BehaviorExperiment, ImagingExperiment)


class ExperimentGroup(Group):

    """Class for representing groups of Experiment objects. These can be
    instantiated directly from an iterable of experiment objects, or via
    helper methods.

    Parameters
    ----------
    iterable_of_expts : iterable of _BaseExperiment instances
        Experiments to include in group
    name : str, optional
        Name for ExperimentGroup object
    parallelize : bool, optional
        Whether to parallelize `apply` calls, defaults to False.

    Examples
    --------
    From experiment objects:

    >>> list_of_expts = [BehaviorExperiment(trial_id=1),
                         BehaviorExperiment(trial_id=2)]
    >>> expt_grp = ExperimentGroup(list_of_expts)

    From trial_ids via helper method:

    >>> expt_grp = ExperimentGroup.from_trial_ids(
            [1, 2], expt_class=BehaviorExperiment
        )

    From database query via helper method:

    >>> expt_grp = ExperimentGroup.from_database(
            project_name='world_domination',
            mouse_name='Pinky'
        )

    An ExperimentGroup instance can be transformed into a Cohort, which is a
    group of ExperimentGroups that organizes the experiments by mouse:

    >>> cohort = expt_grp.to_cohort()
    >>> cohort[0]  # this returns a Mouse object (an ExperimentGroup)
    """

    def __init__(self, iterable_of_expts, **kwargs):
        assert all([isinstance(i, _BaseExperiment)
                    for i in iterable_of_expts]), \
            "All elements must be an instance of an experiment class!"
        super().__init__(iterable_of_expts, **kwargs)

    @classmethod
    def from_trial_ids(cls, trial_ids, expt_class=ImagingExperiment,
                       name='unnamed', parallelize=False, **kwargs):
        """Helper method for constructing ExperimentGroup objects from a list
        of database trial IDs.

        Parameters
        ----------
        trial_ids : list of ints
            Integer database trial_id for each experiment to include in group
        expt_class : _BaseExperiment, optional
            Type of Experiment class to generate for each element of the
            ExperimentGroup. Must be a subclass of _BaseExperiment. Defaults to
            ImagingExperiment
        name : str, optional
            Name for ExperimentGroup object
        parallelize : bool, optional
            Whether to parallelize `apply` calls, defaults to False.
        **kwargs : keyword arguments passed to expt_class initializer
        """
        return cls([expt_class(tid, **kwargs) for tid in trial_ids], name=name,
                   parallelize=parallelize, item_class=expt_class)

    @classmethod
    def from_database(cls, expt_class=ImagingExperiment,
                      name='unnamed', parallelize=False, **db_kwargs):
        """Helper method for constructing ExperimentGroup objects by querying
        the database for the desired experiments.

        Parameters
        ----------
        project_name, mouse_name, experimentType : str, optional
            Return experiments with these property names. Arguments left as
            None (default) are not used to filter.
        expt_class : _BaseExperiment, optional
            Type of Experiment class to generate for each element of the
            ExperimentGroup. Must be a subclass of _BaseExperiment. Defaults to
            ImagingExperiment
        name : str, optional
            Name for ExperimentGroup object
        parallelize : bool, optional
            Whether to parallelize `apply` calls, defaults to False.
        **db_kwargs
            Additional keyword arguments are passed directly to
            lab3.experiment.base.fetch_trials.

        See also
        --------
        lab3.experiment.base.fetch_trials
        """
        trial_ids = fetch_trials(**db_kwargs)
        return cls.from_trial_ids(trial_ids, expt_class=expt_class,
                                  name=name, parallelize=parallelize)

    @property
    def trial_ids(self):
        return [e.trial_id for e in self]

    def to_cohort(self, name=None):
        """Convert this ExperimentGroup into a Cohort (a group of Mice -- so a
        group of ExperimentGroups)."""

        mouse_dict = defaultdict(list)
        for e in self:
            mouse_dict[e.mouse_id].append(e)
        mice = [Mouse(expts, item_class=self.item_class,
                      parallelize=self.parallelize)
                for mouse_id, expts in mouse_dict.items()]

        if name is None:
            name = self.name

        return Cohort(mice, name=name)

    def to_list(self):
        return [e for e in self]

    def __add__(self, other):
        return self.__class__([e for e in self] + [e for e in other],
                              name=self.name, parallelize=self.parallelize)


class Mouse(ExperimentGroup):

    """Mouse objects are ExperimentGroup objects that contain experiments
    from the same mouse.

    Parameters
    ----------
    iterable_of_expts : iterable of _BaseExperiment instances
        Experiments to include in group
    name : ignored
        The group name is automatically set to be the mouse name.
    parallelize : bool, optional
        Whether to parallelize `apply` calls, defaults to False.

    Examples
    --------
    # load all experiments for Pinky

    >>> m = Mouse.from_database('Pinky', experiment_type='random_foraging')

    Note the default experiment class used by Mouse (and ExperimentGroup) is
    ImagingExperiment, so if you haven't paired all experiments with sima_paths
    yet, you should first load the Mouse using BehaviorExperiment:

    >>> from lab3 import BehaviorExperiment
    >>> m = Mouse.from_database('Pinky', expt_class=BehaviorExperiment)

    The Mouse class is endowed with a convenience method to pair the behavior
    experiments with their corresponding imaging datasets

    >>> m = m.pair_imaging_data(imaging_path="/data9/world_domination")
    >>> m[0]  # this is now an ImagingExperiment (if pairing was successful)

    See also
    --------
    lab3.experiment.group.ExperimentGroup
    """

    def __init__(self, iterable_of_expts, **kwargs):

        mouse_ids = [e.mouse_id for e in iterable_of_expts]
        assert len(set(mouse_ids)) == 1, \
            "All experiments must have the same mouse_id!"

        # set the name (or should we use the ID number instead?)
        kwargs['name'] = iterable_of_expts[0].mouse_name
        super().__init__(iterable_of_expts, **kwargs)

    @classmethod
    def from_database(cls, mouse_name, expt_class=ImagingExperiment,
                      parallelize=False, **db_kwargs):
        """Helper method for constructing Mouse objects by querying the
        database for the desired experiments.

        Parameters
        ----------
        mouse_name : str
            Mouse to return
        project_name, experimentType : str, optional
            Return experiments with these property names. Arguments left as
            None (default) are not used to filter.
        expt_class : _BaseExperiment, optional
            Type of Experiment class to generate for each element of the
            ExperimentGroup. Must be a subclass of _BaseExperiment. Defaults to
            ImagingExperiment
        name : ignored
        parallelize : bool, optional
            Whether to parallelize `apply` calls, defaults to False.
        **db_kwargs
            Additional keyword arguments are passed directly to
            lab3.experiment.base.fetch_trials.

        See also
        --------
        lab3.experiment.base.fetch_trials
        lab3.experiment.group.ExperimentGroup
        """
        trial_ids = fetch_trials(mouse_name=mouse_name, **db_kwargs)
        return cls.from_trial_ids(trial_ids, expt_class=expt_class,
                                  parallelize=parallelize)

    def pair_imaging_data(self, *args, do_nothing=False, **kwargs):
        """Use the Hungarian algorithm to find the optimal pairings between
        behavior experiments and imaging datasets, using the start time
        difference between the behavior tdml and Prairie xml files as the cost
        function.

        Parameters
        ----------
        imaging_path : str
            Locate Prairie xmls/sima folders below this folder
        max_dt : float, optional
            Maximum timestamp difference (in seconds) between behavior tdml and
            Prairie xml files. Defaults to 60
        force_pairing : bool, optional
            Whether to overwrite existing sima pairs. Defaults to False
        keywords : iterable of strs, optional
            Additional keywords to identify in sima_paths
        do_nothing : bool, optional
            Find pairs but don't update the database. Defaults to False
        verbose : bool, optional
            Display extra information. Defaults to False
        """

        assert all([isinstance(e, BehaviorExperiment) for e in self]), \
            "Cannot find imaging/behavior pairs unless all members of " \
            + "ExperimentGroup are instances of BehaviorExperiment"

        from lab3.experiment._pairing import pair_mouse_experiments
        pair_mouse_experiments(self, *args, do_nothing=do_nothing, **kwargs)

        if not do_nothing:
            # Note this may still break if some experiments were unpaired
            return Mouse.from_trial_ids([e.trial_id for e in self],
                                        expt_class=ImagingExperiment,
                                        parallelize=self.parallelize)
        else:
            return self


class Cohort(Group):

    """Cohort objects are groups of Mouse objects.

    Examples
    --------
    >>> cohort[0]  # this is a Mouse (ExperimentGroup)
    >>> cohort[0][0]  # this is an Experiment object
    """

    def __init__(self, iterable_of_mice, **kwargs):
        assert all([isinstance(m, Mouse) for m in iterable_of_mice]), \
            "All elements must be an instance of the Mouse class!"
        super().__init__(iterable_of_mice, **kwargs)

    @classmethod
    def from_ExperimentGroup(cls, expt_grp, name='unnamed'):
        return expt_grp.to_cohort(name=name)

    @classmethod
    def from_database(cls, name='unnamed', **kwargs):
        expt_grp = ExperimentGroup.from_database(**kwargs)
        return cls.from_ExperimentGroup(expt_grp, name=name)

    @property
    def trial_ids(self):
        trial_ids = []
        for grp in self:
            trial_ids.extend(grp.trial_ids)
        return trial_ids

    def to_list(self):
        return reduce(lambda x, y: x + y, [grp.to_list() for grp in self])


class MouseGroup(Cohort):

    """This is simply an alias for Cohort"""

    pass


class CohortGroup(Group):

    """CohortGroup objects are groups of Cohorts.

    Examples
    --------
    >>> cohort[0]  # this is a Cohort
    >>> cohort[0][0]  # this is a Mouse (ExperimentGroup)
    >>> cohort[0][0][0]  # this is an Experiment object
    """

    def __init__(self, iterable_of_cohorts, **kwargs):
        assert all([isinstance(m, Cohort) for m in iterable_of_cohorts]), \
            "All elements must be an instance of the Cohort class!"
        super().__init__(iterable_of_cohorts, **kwargs)

    @property
    def trial_ids(self):
        trial_ids = []
        for grp in self:
            trial_ids.extend(grp.trial_ids)
        return trial_ids

    def to_list(self):
        return reduce(lambda x, y: x + y, [grp.to_list() for grp in self])



# import lab3
# import lab3.analysis.basic as lab
import lab3.experiment as experiment
# import lab3.signal as signal
import inspect

# we should only include classes here that subclass Item...?

experiment_classes = dict(inspect.getmembers(experiment.base, inspect.isclass))
# event_classes = dict(inspect.getmembers(signal.events, inspect.isclass))
group_classes = dict(inspect.getmembers(experiment.group, inspect.isclass))

REGISTRY = {}
REGISTRY.update(experiment_classes)
# REGISTRY.update(event_classes)
REGISTRY.update(group_classes)


def lookup_class(name):

    try:
        return REGISTRY[name]
    except Exception:
        return REGISTRY['Item']



import pandas as pd
import os
from lab3.core.metadata import Metadata, MetadataGroup

class FileMetadata(Metadata):
    constructor_ids = ['file_index_common', 'df_index_common', 'metadata_columns', 'file_name']

    def bind(self, expt):
        return pd.MultiIndex.from_frame(
            self.read_file_data(expt)[self.metadata_columns])

    def bind_to(self, signals, expt):
        filedata = self.read_file_data(expt)
        filedata[self.df_index_common] = filedata[self.file_index_common]
        filedata = filedata.set_index(self.df_index_common)

        signals_index = signals.reset_index()[self.df_index_common].iloc[0]
        
        try:
            signals_index.shape
            index_type = [type(i) for i in signals_index]
        except:
            index_type = type(signals_index)

        filedata.index = filedata.index.astype(index_type)

        #dtype = signals.reset_index()[self.df_index_common].dtype

        return signals.join(filedata[self.metadata_columns],
                            on=[self.df_index_common]).set_index(
                            self.metadata_columns, append=True)

    def read_file_data(self, expt):
        raise NotImplementedError

class ExcelMetadata(FileMetadata):
    def read_file_data(self, expt):
        filedata = pd.read_excel(os.path.join(expt.sima_path, self.file_name))
        return filedata

class CSVMetadata(FileMetadata):
    def read_file_data(self, expt):
        filedata = pd.read_csv(os.path.join(expt.sima_path, self.file_name))
        return filedata






import pandas as pd
import numpy as np
import copy
import scipy.signal as ss

from scipy.ndimage.filters import median_filter

from lab3.core import Item, Group
import lab3.experiment.base as leb

from .base import EventsFile, Event, OnsetEvent, OffsetEvent, IntervalEvent, bool2events

class Group2DataFrameMixin:
    @classmethod
    def load(cls, events_file_path, event_type, channel=None, label=None, trigger=None):
        # Handle object of type `Experiment` passed
        try:
            events_file_path = events_file_path.imaging_events_path
        except AttributeError:
            pass

        with EventsFile(events_file_path) as events_file:
            events = events_file.get(f"{label}/{event_type}/{channel}")

        return cls.from_pandas(events, name=event_type, trigger=trigger)

    def save(self, events_file_path, event_type, channel=None, 
             label=None):
        with EventsFile(events_file_path) as events_file:
            events_file.put(f"{label}/{event_type}/{channel}", self.dataframe)        

    @classmethod
    def from_pandas(cls, dataframe, name='unnamed', item_class=None, **kwargs):
        if item_class is None:
            item_class = cls.item_class
        return cls([item_class(**row, **kwargs) for _, row in dataframe.iterrows()], 
                    name=name, item_class=item_class)

    def to_pandas(self):
        return pd.concat([ev.to_pandas() for ev in self],
                keys=pd.RangeIndex(len(self)), 
                names=[self.item_class.name]).unstack()        

    @property
    def dataframe(self):
        return self.to_pandas()        

class EventGroup(Group, Group2DataFrameMixin):

    item_class = Event
    
    @property
    def onset(self):
        return pd.Series([event.onset for event in self], name='onset')
    
    @property
    def offset(self):
        return pd.Series([event.offset for event in self], name='offset')
    
    def expand(self, pre=0, post=0):
        df = self.dataframe
        df['onset'] -= pre
        df['offset'] += post
        return self.from_pandas(df)

    def frames2times(self, expt):
        new_events = []
        imaging_times = pd.Series(expt.imaging_times())
        for event in self:
            ids = {
                key:imaging_times.loc[value] 
                    for key, value in event.constructors().items()
            }
            new_events.append(self.item_class(**ids))

        return self.__class__(new_events)
    
    def bool_array(self, n_frames):
        bool_array = np.zeros(n_frames)
        idx_array = np.arange(n_frames)
        for event in self:
            bool_array[(idx_array < event.offset) & (event.onset <= idx_array)] = 1

        return bool_array.astype(bool)
    
    def __contains__(self, time):
        return np.any([time in event for event in self])

    def filter_trials(self, trial_filter):
        return self.from_pandas(self.dataframe.loc[trial_filter], name=self.name, 
                                item_class=self.item_class)

class LFPEventGroup(EventGroup):
    item_name = 'LFPEvent'
    @classmethod
    def load(cls, expt, event_type, label='LFP', channel=None):
        return super().load(expt.lfp_events_path, event_type, 
                                label=label, channel=channel)
    
class BehaviorEventGroup(EventGroup):
    item_name = 'BehaviorEvent'
    @classmethod
    def load(cls, expt, behavior_key, as_time_events=False, **kwargs):
        if behavior_key == 'running':
            events = cls._load_running(expt, **kwargs)
        elif behavior_key == 'lap':
            events = cls._load_lap(expt, **kwargs)
        elif behavior_key == 'position':
            events = cls._load_position(expt, **kwargs)
        else:
            events = cls._load_generic(expt, behavior_key, **kwargs)

        if as_time_events:
            return events.frames2times(expt)
        else:
            return events

    @classmethod
    def _load_running(cls, expt, threshold=4, debounce_window=4, fmt='interval'):
        behavior_data = expt.format_behavior_data()
        velocity = behavior_data['velocity'].astype(int)
        # window must be odd
        try:
            window = 2*(int(expt.frame_rate*debounce_window)//2)+1
        except AttributeError:
            window = 2*(int(10*debounce_window)//2)+1
        running_array = median_filter(velocity > threshold, window).astype(int)

        if fmt == 'offset':
            events = cls(bool2events(running_array, event_class=OffsetEvent), name='running_start')
        elif fmt == 'onset':
            events = cls(bool2events(running_array, event_class=OnsetEvent), name='running_stop')
        else:
            events = cls(bool2events(running_array, event_class=IntervalEvent), name='running')

        return events

    @classmethod
    def _load_position(cls, expt, position):
        behavior_data = expt.format_behavior_data()
        position_array = behavior_data['treadmillPosition'] - position
        onsets, = np.where(np.diff(np.signbit(position_array)))

        events = cls([OnsetEvent(trigger) for trigger in onsets], 
                    name=f"position_{position:0.3f}")

        return events

    @classmethod 
    def _load_lap(cls, expt, **kwargs):
        from .lap import LapGroup
        return LapGroup.from_expt(expt)

    @classmethod
    def _load_generic(cls, expt, behavior_key, fmt='interval'):
        behavior_data = expt.format_behavior_data()
        behavior_array = behavior_data[behavior_key].astype(int)
        
        if fmt == 'offset':
            events = cls(bool2events(behavior_array, event_class=OffsetEvent), name=behavior_key)
        if fmt == 'onset':
            events = cls(bool2events(behavior_array, event_class=OnsetEvent), name=behavior_key)
        else:
            events = cls(bool2events(behavior_array, event_class=IntervalEvent), name=behavior_key)

        return events

class EventGroupGroup(Group):
    item_class = EventGroup

    def to_pandas(self):
        return pd.concat([grp.dataframe for grp in self],
                keys=[grp.name for grp in self],
                names=[self.item_class.name])

    @property
    def dataframe(self):
        return self.to_pandas()

class ImagingEventGroupGroup(EventGroupGroup, Group2DataFrameMixin):
    def to_pandas(self):
        return pd.concat([grp.dataframe for grp in self],
                keys=[grp.name for grp in self],
                names=['roi_label'])




import numpy as np
import itertools as it

from sima.sequence import _Sequence_HDF5


class Struct(object):

    def __init__(self, **entries):
        self.__dict__.update(entries)


def _pool_helper(args):
    return args[0](*args[1:])


def pool_helper(pool, method, *args):
    args = list(args)

    for i, _ in enumerate(args):
        if not hasattr(args[i], '__iter__'):
            args[i] = it.repeat(args[i])

    if pool is None:
        return map(_pool_helper, zip(it.repeat(method), *args))

    return pool.map(_pool_helper, zip(it.repeat(method), *args))


def frame_blocks(frames, chunk_size):
    try:
        t1, t2 = frames
    except TypeError:
        t1 = 0
        t2 = frames

    if chunk_size == 0:
        return [(t1, t2)]

    n_frames = t2-t1
    blocks = np.linspace(
        t1, t2, int(np.ceil(float(n_frames)/chunk_size)+1),
        dtype=int)
    return list(zip(blocks[:-1], blocks[1:]))


def slice_sequence(seq, n_processes=4, max_len=None):
    n_processes = max(1, n_processes-1)

    if max_len is None:
        max_len = np.ceil(seq.shape[0]/n_processes)

    batches = frame_blocks(seq.shape[0],
                           min(max_len, np.ceil(seq.shape[0]/n_processes)))

    return map(
        lambda seq_indices: seq_indices[0][slice(*seq_indices[1])],
        zip(it.repeat(seq), batches))


def bind_locked_get_frame(seq, lock):
    def locked_get_frame(self, t):

        """Get the frame at time t, but not clipped"""
        slices = tuple(slice(None) for _ in range(self._T_DIM)) + (t,)

        with lock:
            for i in range(11):
                try:
                    frame = self._dataset[slices]
                    break
                except Exception:
                    pass

        if i == 10:
            raise Exception

        swapper = [None for _ in range(frame.ndim)]
        for i, v in [(self._Z_DIM, 0), (self._Y_DIM, 1),
                     (self._X_DIM, 2), (self._C_DIM, 3)]:
            if i >= 0:
                j = i if self._T_DIM > i else i - 1
                swapper[j] = v
            else:
                swapper.append(v)
                frame = np.expand_dims(frame, -1)
        assert not any(s is None for s in swapper)
        for i in range(frame.ndim):
            idx = swapper.index(i)
            if idx != i:
                swapper[i], swapper[idx] = swapper[idx], swapper[i]
                frame = frame.swapaxes(i, idx)
        assert swapper == [0, 1, 2, 3]
        assert frame.ndim == 4
        return frame.astype(float)

    h5_seq = seq
    while type(h5_seq) != _Sequence_HDF5:
        h5_seq = h5_seq._base

    bound = locked_get_frame.__get__(h5_seq, h5_seq.__class__)
    setattr(h5_seq, '_get_frame', bound)

    # This is for the _InterpolateGapFilledSequence Sequence wrapper in sima
    # TODO: is there a better way to do this?
    try:
        seq._unmasked_base
    except AttributeError:
        pass
    else:
        bind_locked_get_frame(seq._unmasked_base, lock)



import numpy as np
import pandas as pd
from scipy.ndimage import correlate1d

from .abstract_spatial_tuning import SpatialTuning

def _cyclic_autocorr(array, mode='wrap', pearson=True):
    output = np.zeros(len(array) + 1, dtype=float)
    
    if pearson:
        array -= np.mean(array)
        array /= np.std(array)*np.sqrt(len(array))
    
    correlate1d(array, array, mode=mode, output=output[:-1], origin=0)
    output[-1] = output[0]
    
    return output

class AutocorrelationMixin:
    
    constructor_ids = ['max_lag']
    
    def to_columns(self):
        return pd.RangeIndex(-self.max_lag, self.max_lag+1)
    
    def calculate(self, dataframe, **acorr_kwargs):
        autocorrs = dataframe.apply(_cyclic_autocorr, **acorr_kwargs, axis=1)
        autocorrs = autocorrs.apply(pd.Series)
        autocorrs.columns = self.columns
        return autocorrs        

class SpatialAutocorrelation(AutocorrelationMixin, SpatialTuning):
    
    constructor_ids = SpatialTuning.constructor_ids + AutocorrelationMixin.constructor_ids
    
    def __init__(self, **spatial_tuning_kwargs):
        self.spatial_tuning = SpatialTuning(**spatial_tuning_kwargs)
        super().__init__(**spatial_tuning_kwargs)
        self.max_lag = self.nbins//2
    
    def apply_to(self, expt, signal_spec):
        tuning_curves = self.spatial_tuning.apply_to(expt, signal_spec=signal_spec)
        return self.calculate(tuning_curves)
    
class SpatialAutocorrelationWidth(SpatialAutocorrelation):
    
    constructor_ids = SpatialAutocorrelation.constructor_ids + ['cutoff']
    
    def __init__(self, cutoff, **spatial_tuning_kwargs):
        self.spatial_autocorrelation = SpatialAutocorrelation(**spatial_tuning_kwargs)
        self.cutoff = cutoff
        super().__init__(**spatial_tuning_kwargs)
        
    def to_columns(self):
        return pd.Index(["width"])
    
    def apply_to(self, expt, signal_spec, groupby=None, agg=None):
        spatial_autocorrelation = self.spatial_autocorrelation.apply_to(expt, signal_spec=signal_spec)
        
        if groupby is not None and agg is not None:
            spatial_autocorrelation = spatial_autocorrelation.groupby(groupby).agg(agg)
            
        return self.calculate(spatial_autocorrelation)
        
    def calculate(self, spatial_autocorrelation):
        def get_width(series):
            if np.any(np.isnan(series)):
                return np.nan
            else:
                return 2*np.argmax(series.loc[0:] < self.cutoff)
        
        widths = spatial_autocorrelation.apply(get_width, axis=1)
        widths = widths.to_frame()
        widths.columns = self.columns
        return widths



"""Core set of classes that form the basis of most objects in the LAB3 module.
"""

from abc import ABCMeta, abstractmethod

import numpy as np
import pandas as pd
import itertools as it
from pandas.core.dtypes.common import is_scalar

from lab3.misc import level_accessor
from lab3.core.helpers import infer_item_class, bind_args, is_target_object, ResultStore

class ConstructorMeta(ABCMeta):

    """Custom metaclass to force the user to set all constructor ids."""

    constructor_ids = []

    def __call__(self, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)

        if not hasattr(obj, 'name'):
            obj.name = obj.__class__.__name__

        for attr_name in obj.constructor_ids:
            try:
                getattr(obj, attr_name)
            except AttributeError:
                raise AttributeError(
                    f'Required constructor attribute ({attr_name}) not set')
        return obj


class Item(metaclass=ConstructorMeta):

    """Abstract base class for modeling single entities. When implementing a
    subclass of Item, you must determine a list of attributes shared by all
    instances of the class, which can be used to uniquely reconstruct each
    item instance. This list is defined as the class attribute
    `constructor_ids`.

    Subclasses of Item cannot be instantiated unless all attributes specified
    in `constructor_ids` are set (i.e. via __init__).

    Examples
    --------
    Create a new Item class:

    >>> class Person(Item):
    >>>     constructor_ids = ['name', 'position']
    >>>     def __init__(self, name, position='grad_student'):
    >>>         self.name = name
    >>>         self.position = position

    Instantiate the item and run an analysis on the object instance

    >>> obj = Person('zhenrui', position='grad_student')
    >>> obj.apply(...)  # pass an instance of an analysis class here
    """

    # List of ids sufficient to reconstruct the item uniquely
    constructor_ids = []

    def __init__(self, **constructors):
        """This will set all keyword arguments as instance attributes. In
        practice, you should implement `__init__` in derived classes"""
        self._register_kwargs(constructors)

    def _register_kwargs(self, kwargs):
        self.__dict__.update(kwargs)

    def constructors(self):
        """Returns a dict of parameters sufficient to uniquely identify
        and instantiate this object
        """
        return {
            cid: self.__dict__[cid] for cid in self.constructor_ids
        }

    def label(self):
        """Returns scalar identifier"""
        # Why should we assume this is just the first constructor value?
        try:
            return list(self.constructors().values())[0]
        except Exception:  # what error should we be catching here?
            return None

    def identifiers(self):
        """Return a dict of identifiers (for information display only)"""
        return self.constructors()

    def apply(self, analysis, *args, groupby=None, agg=None, **kwargs):
        """doc"""
        result = analysis.apply_to(self, *args, **kwargs)

        if groupby is not None and agg is not None:
            return result.groupby(groupby).agg(agg)
        else:
            return result

    def is_target_object(self, of):
        return is_target_object(self, function=of.apply_to, name=of.name)

    @property
    def _df(self):
        return pd.DataFrame(np.zeros((1, 0)))

    def __hash__(self):
        return hash(frozenset(self.constructors().items()))

    def __eq__(self, other):
        return hash(self) == hash(other)

    # def __repr__(self):
    #     return "<{} {}>".format(self.__class__.__name__, ' '.join(
    #         [f'{key}="{value}"' for key, value in self.identifiers().items()]))

    def __repr__(self, line_max=80):
        """TODO : Make this pretty"""
        repr_ = type(self).__name__
        repr_ += '('
        base_length = len(repr_)
        curr_line_length = len(repr_)
        for key, value in self.identifiers().items():
            param_string = key + '=' + str(value) + ', '
            if (curr_line_length + len(param_string)) <= line_max:
                repr_ += param_string
                curr_line_length += len(param_string)
            else:
                repr_ += '\n' + ' ' * base_length
                repr_ += param_string
                curr_line_length = base_length + len(param_string)
        repr_ = repr_[:-2] + ')'
        return repr_

    def __str__(self):
        return repr(self)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)


class Group(list, Item):

    """Class for modeling a collection of Items of homogenous type. A Group is
    also an Item, and so Groups can be collections of Groups, allowing for
    heirarchical organization of objects.

    Examples
    --------
    Assume we have defined an Item subclass `Person`.
    Nesting collections of Items:

    >>> people = [Person('zhenrui'), Person('james'), Person('jack')]
    >>> grp = Group(iterable_of_items=people, name='code_review')
    >>> grp_of_grps = Group([grp] * 2)

    Indexing is straightforward. This returns a group:

    >>> grp_of_grps[0]

    This returns a Person:

    >>> grp_of_grps[0][0]

    Slicing returns a group:

    >>> grp_of_grps[0][0:2]
    >>> grp_of_grps[0:2][0:2]

    Since Groups are Items, we may define analysis classes that operate on
    Groups, and apply these with the same syntax as regular Items:

    >>> grp.apply(...)
    >>> grp_of_grps.apply(...)

    By subclassing Group, you can augment it with functionality specific
    to certain item classes:

    >>> class Meeting(Group):
    >>>     constructor_ids = ['name', 'meeting_type']
    >>>     def __init__(self, people, meeting_type='general', **kwargs):
    >>>         assert all([isinstance(p, Person) for p in people]), \
    >>>             'All elements of "people" must be Person objects!'
    >>>         super().__init__(iterable_of_items=people, **kwargs)
    >>>         self.meeting_type = meeting_type
    >>>
    >>>     def my_custom_method(self):
    >>>         # Here I implement a function specific to collections of
    >>>         # Person objects
    """

    constructor_ids = ['name', 'parallelize']

    def __init__(self, iterable_of_items=(), parallelize=False,
                 item_class=None, name='unnamed', **kwargs):
        super().__init__(iterable_of_items)
        if item_class is None:
            # should this be a class attribute or abstract property?
            self.item_class = infer_item_class(iterable_of_items)
        else:
            self.item_class = item_class
        self.parallelize = parallelize
        self.name = name

        self._register_kwargs(kwargs)

    @property
    def _df(self):
        return pd.concat([self._box(item) for item in self])

    def apply(self, analysis, *args, groupby=None, agg=None, **kwargs):
        """By default no groupby"""
        return AnalysisResult(self, analysis, *args, groupby=groupby, agg=agg, **kwargs)
    @property
    def index(self):
        return self._df.index

    @classmethod
    def from_index(cls, index, name=None):
        frame = index.to_frame(index=False)
        return cls.from_frame(frame, name=name)

    @classmethod
    def from_frame(cls, frame, name=None):
        # This MUST stay here or a circular import will result!
        from lab3.core.registry import lookup_class

        level_name = frame.columns[0]
        if len(frame.columns) > 1:
            return Group([Group.from_frame(df.drop(level_name, axis=1),
                                           name=group_name)
                          for group_name, df in frame.groupby(level_name)],
                         name=name, item_class=lookup_class(level_name))
        else:
            return Group.from_scalars(frame[level_name], name=name,
                                      item_class=lookup_class(level_name))

    def level(self, *args, **kwargs):
        return self.from_frame(self._df.level(*args, **kwargs).reset_index(),
                               name=self.name)

    @classmethod
    def from_scalars(cls, scalars, item_class, name=None, **kwargs):
        # This only works if item_class only has one position arg to be set!
        return cls([item_class(x) for x in scalars], name=name, **kwargs)

    def _box(self, item):
        boxed_df = pd.concat([item._df], keys=[item.label()],
                             names=[self.item_class.__name__])
        if None in boxed_df.index.names:
            boxed_df.index = boxed_df.index.droplevel(None)

        return boxed_df

    def __add__(self, other):
        return self.__class__(self + other)

    def __getitem__(self, key):
        if is_scalar(key):
            return super().__getitem__(key)
        else:
            return self.__class__(super().__getitem__(key))

    def __repr__(self):
        return self.index.to_frame(index=None).__repr__()

    def _repr_html_(self):
        return f"<div><b>{self.__class__.__name__}: " \
               + f"'{self.label()}'</b></div>" + self._df._repr_html_()


class Analysis(Item):

    """Abstract base class for designing analyses."""

    @abstractmethod
    def to_columns(self):
        """Define a priori what columns are expected in the output.
        This should NOT take any parameters but only use what is available
        as an instance attribute.

        Returns
        -------
        pd.Index
        """

    @abstractmethod
    def apply_to(self, item, *args, **kwargs):
        """Specifies how to go from an Item (Event, Experiment, Mouse, Group,
        etc.) to results.

        Should load the data from item, preprocess *args and **kwargs,
        and call self.calculate

        TODO: Maybe this can be formalized into abstract methods
        """

    @abstractmethod
    def calculate(self, data, *args, **kwargs):
        """Implement the analysis algorithm here. Unlike `apply`, this should
        take a dataframe and return a dataframe.
        """

    @property
    def columns(self):
        return self.to_columns()

    @property
    def _df(self):
        return pd.DataFrame([], columns=self.columns)

    def get_params(self):
        params = {}
        for key, val in self.__dict__.items():
            if key[0] != '_':
                params[key] = val
        return params

    # def __repr__(self, line_max=80):
    #     """TODO : Make this pretty"""
    #     repr_ = type(self).__name__
    #     repr_ += '('
    #     base_length = len(repr_)
    #     curr_line_length = len(repr_)
    #     for key, value in self.get_params().items():
    #         param_string = key + '=' + str(value) + ', '
    #         if (curr_line_length + len(param_string)) <= line_max:
    #             repr_ += param_string
    #             curr_line_length += len(param_string)
    #         else:
    #             repr_ += '\n' + ' ' * base_length
    #             repr_ += param_string
    #             curr_line_length = base_length + len(param_string)
    #     repr_ = repr_[:-2] + ')'
    #     return repr_

    def _repr_html_(self):
        return pd.DataFrame([], columns=self.columns)._repr_html_()


    def save(self, df, expt, verbose=False, **runtime_parameters):        
        parameters = {"analysis": self, "runtime_parameters": runtime_parameters}
        
        with ResultStore(expt.sima_path) as f:
            f.put(parameters, df)
            
        message = f'Saving results of analysis: {type(self).__name__} \n on experiment {expt}'
        if self.verbose:
            print(message)
    
    def load(self, expt, verbose=False, **runtime_parameters):        
        parameters = {"analysis": self, "runtime_parameters": runtime_parameters}
        
        with ResultStore(expt.sima_path) as f:
            data = f.get(parameters)
            
        self.loaded = True
        
        message = f'Loaded results of analysis: {type(self).__name__} \n on experiment {expt}'
        if self.verbose:
            print(message)
            
        return data

class Automorphism(Analysis):

    """Abstract parent class for analyses which are automorphisms, i.e.,
    they take a dataframe and return a dataframe of the same shape.

    As a result, columns may be unknown a priori. A consequence of this is that
    these analyses cannot necessarily be joined when mapped across a Group
    """
    def to_columns(self):
        return pd.Index(["__IDEM__"])

class AnalysisGroup(Analysis, Group): 
    """Join multiple analyses together. Allows easy creation of matrices.
    """
    def to_columns(self):
        cols = [pd.concat([analysis._df], keys=[analysis.name], 
                        names=['Analysis', 'Feature'], axis=1)
                    for analysis in self]
        
        return pd.concat(cols, axis=1).columns
    
    def apply_to(self, item, *args, args_list=(), kwargs_list=(), 
              **kwargs):
        try:
            args_list[0][0]
        except IndexError:
            args_list = it.repeat(args)
        try:
            kwargs_list[0]
        except IndexError:
            kwargs_list = it.repeat(kwargs)
        
        results = [analysis.apply_to(item, *args, **kwargs) 
                       for analysis, args, kwargs in zip(self, args_list, kwargs_list)]
        try:
            result = pd.concat(results, axis=1)
            result.columns = self.columns
            return result
        except:
            return results

class AnalysisResult(pd.DataFrame):

    _metadata = ['group', 'analysis', 'error_policy', '_groupby', '_agg']

    def __init__(self, group=None, analysis=None, data=None, *args, groupby=None, agg=None, 
                 error_policy='warn', **kwargs):

        if analysis is not None:
            self.group = group 
            self.analysis = analysis 
            self._groupby = groupby
            self._agg = agg
            self.error_policy = error_policy
            
            # TODO: Maybe this can happen lazily/later
            # TODO: Maybe args and kwargs shouldn't get passed here
    #        super().__init__(index=group, columns=analysis)

            data = self.calculate(*args, **kwargs)
            super().__init__(data)
        elif data is not None:
            super().__init__(data=data, **kwargs)
        else:
            # Fall-through case
            super().__init__(group, **kwargs)
        
    @property
    def _constructor(self):
        return AnalysisResult

    def calculate(self, *args, **kwargs):
        """By default no groupby
        """
        if self.group.parallelize is True:
            map_df = self._map_parallel(self.analysis, *args, **kwargs)
        else:
            map_df = self._map(self.analysis, *args, **kwargs)

        if self._groupby and self._agg:
            result_df = map_df.groupby(self._groupby).agg(self._agg)
        else:
            result_df = map_df

        return result_df
    
    def _map(self, analysis, *args, **kwargs):
        results = []
        for item in self.group:
            try:
                if item.is_target_object(of=analysis):
                    results.append(analysis.apply_to(item, *args, **kwargs))
                else:
                    results.append(item.apply(analysis, *args, **kwargs))
            except Exception as exc:
                if self.error_policy == 'ignore':
                    #pass
                    results.append(None)
                elif self.error_policy == 'warn':
                    print(f"Failed on {item} with {exc}!")
                    results.append(None)
                else:
                    print(f"Failed on {item} with {exc}!")
                    raise exc                    

        return self._join_results(results)

    def _map_parallel(self, analysis, *args, n_processes=8, **kwargs):
        try:
            from multiprocess import Pool
        except ImportError:
            import warnings
            warnings.warn("It seems you do not have `multiprocess` installed, falling back to plain `multiprocessing`!"
                    " multiprocess works better for parallelization")
            from multiprocessing import Pool
        from functools import partial

        # TODO: Just noticed this behavior is different from _map, probably incorrect
        # This only allows parallelization across the fundamental Item of the analysis

        # TODO: error handling

        apply_analysis = partial(analysis.apply_to, *args, **kwargs)

        with Pool(n_processes) as p:
            results = p.map(apply_analysis, self.group)

        return self._join_results(results)

    def _join_results(self, results):
        keys = [item.label() for item in self.group]
        name = self.group.item_class.__name__

        new_results = []
        new_keys = []

        for k, r in zip(keys, results):
            if r is not None and not r.empty:
                new_results.append(r)
                new_keys.append(k)

        try:
            aligned_results = [new_results[0].align(r, axis=0)[1].dropna(how='all') for r in new_results]
            return pd.concat(aligned_results, keys=new_keys, names=[name])
            #return pd.concat([r for r in results if not r.empty], keys=keys, names=[name])
        except ValueError:
            return None
        except Exception as exc:
            print(f"Joining results failed due to {exc}")
            raise
            #try:
            #    return [pd.concat([res], keys=[key], names=[name])
            #        for res, key in zip(results, keys)]
            #except:
            #    return dict(zip(keys, results))



"""Dedicated ImagingDataset classes for extracting and importing signals via
Suite2p"""

import os
import shutil
import warnings

# import before numpy so we set the environment variables
from lab3.extraction.s2p import settings, extract_signals, import_signals
from lab3.misc.sima_sequences import _Suite2pRigidSequence
# from lab3.misc.sima_helpers import backup_dir, get_base_sequence
from lab3.misc.sima_compatibility import sima_compatible

import numpy as np
from abc import abstractmethod, ABCMeta
from sima import ImagingDataset
from pkg_resources import get_distribution


class Suite2pStrategy(ImagingDataset, metaclass=ABCMeta):

    """Abstract class for creating Suite2p strategies. We inherit from
    sima.ImagingDataset, in order to interface with the underlying
    imaging data. Accordingly, instances should be generated via the load
    method and not instantiated directly (i.e. we assume a sima folder
    already exists).

    Implement :meth:`__init__`, :meth:`extract`, and :meth:`import_results` to
    subclass.

    See also
    --------
    sima.ImagingDataset
    """

    def __str__(self):
        return f'<{type(self).__name__}>'

    def __repr__(self):
        return ('<Suite2pImagingDataset: '
                + f'num_sequences={self.num_sequences},'
                + f'frame_shape={self.frame_shape}, '
                + f'num_frames={self.num_frames}>')

    @property
    def s2p_path(self):
        return os.path.join(self.savedir, "suite2p")

    @property
    def ops_list_path(self):
        return os.path.join(self.s2p_path, "ops1.npy")

    @property
    def ops(self):
        return self._ops

    @ops.setter
    def ops(self, user_ops):
        """Creates the ops dictionary for Suite2p input

        Parameters
        ----------
        user_ops : dict
            Pass keyword/argument pairs as a dictionary to override
            the default Suite2p settings. See OPS for option details.
        """

        user_ops["xrange"] = np.array([0, self.frame_shape[2]])
        user_ops["yrange"] = np.array([0, self.frame_shape[1]])
        self._ops = {**settings.get_ops(
            get_distribution("suite2p").version),
                     **user_ops}

    @property
    def ops_list(self):
        try:
            return np.load(self.ops_list_path, allow_pickle=True)
        except FileNotFoundError:
            raise AttributeError("ops_list does not exist. Call extract to "
                                 + "process this dataset first.")

    @property
    def db(self):
        """Creates the db dictionary for Suite2p input"""
        # merge with default db
        # TODO: There should be a way to get this to work, looks like
        # there are some 'under the hood' issues with shutil
        # 'fast_disk': os.path.join("/fastscratch/scratch/suite2p/fast_disk",
        #    os.path.basename(self.savedir))}}
        return {**settings.DB, **{'data_path': [self.savedir],
                                  'save_path0': self.savedir,
                                  'fast_disk': self.savedir}}

    @property
    def bad_frames_path(self):
        return os.path.join(self.savedir, "bad_frames.npy")

    @property
    def bad_frames(self):
        """Frames to ignore during processing"""
        return np.load(self.bad_frames_path)

    @property
    def results_path(self):
        """For multiplane data assume the user has curated the combined view"""
        subfolder = "combined" if os.path.exists(
            os.path.join(self.s2p_path, "combined")) else "plane0"
        return os.path.join(self.s2p_path, subfolder)

    @property
    def raw_signals(self):
        """Returns the raw fluorescence traces for each ROI"""
        return np.load(os.path.join(self.results_path, "F.npy"),
                       allow_pickle=True)

    @property
    def npil_signals(self):
        """Returns the neuropil fluorescence traces for each ROI"""
        return np.load(os.path.join(self.results_path, "Fneu.npy"),
                       allow_pickle=True)

    @property
    def cell_indicator(self):
        """Returns a list of row indices corresponding to accepted cells"""
        try:
            return self._cell_indicator
        except AttributeError:
            self._cell_indicator = np.where(
                np.load(os.path.join(self.results_path, "iscell.npy"),
                        allow_pickle=True)[:, 0])[0]
            return self._cell_indicator

    def _modify_cell_indicator(self, value):
        assert value in [0, 1], "value must be 0 (False) or 1 (True)!"
        try:
            del self._cell_indicator
        except AttributeError:
            pass
        ci = np.load(os.path.join(self.results_path, "iscell.npy"),
                     allow_pickle=True)
        ci[:, 0] = value
        np.save(os.path.join(self.results_path, "iscell.npy"), ci)

    def accept_all_cells(self):
        """Classify all ROIs as cells"""
        self._modify_cell_indicator(1)

    def reject_all_cells(self):
        """Classify all ROIs as not cells"""
        self._modify_cell_indicator(0)

    @property
    def static_indicator(self):
        """Returns a list of row indices corresponding to cells with the
        static marker"""
        raise NotImplementedError

    def get_stat_file(self, plane=0):
        """Returns list of dictionaries with additional ROI information"""
        return np.load(os.path.join(
            self.s2p_path, f"plane{plane}", "stat.npy"), allow_pickle=True)

    # ----------------- Implement these methods to subclass ----------------- #

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def extract(self):
        pass

    @abstractmethod
    def import_results(self):
        pass

    # ------------------------------- Optional ------------------------------ #

    def apply_mc_results(self):
        pass


@sima_compatible
class Suite2pImagingDataset(Suite2pStrategy):

    """The main interface for using Suite2p with SIMA. Suite2pImagingDatasets
    are instantiated via the load method, inherited from ImagingDataset. They
    are not initialized directly.

    Examples
    --------
    >>> from lab3.extraction.s2p import Suite2pImagingDataset
    >>> ds = Suite2pImagingDataset.load("/path/to/my/folder.sima")
    >>> ds.extract()

    Following extraction, the user must manually curate the results in the
    Suite2p GUI on their local machine. Signals can then be imported using
    the object again:

    >>> ds.import_results()
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract(self, overwrite=False, reprocess=False, **kwargs):
        """Run Suite2p extraction on imaging dataset.

        Prior to conversion, the imaging dataset is converted to a binary file,
        in the format expected by Suite2p. Along the way, additional 'ops'
        dictionaries are created for each imaging plane and saved to disk in
        the suite2p directory.

        If the dataset has multiple sequences, they will be concatenated in the
        binary file as if they were one continuous recording. For this reason,
        all sequences should be of compatible frame shape (zyxc). This allows
        batch processing of datasets from the same FOV, but special
        consideration should be given for motion correction. If the individual
        datasets have already been motion corrected, you must align the
        sequences and crop them accordingly when constructing the batch imaging
        dataset object. This step is unnecessary if using Suite2p's motion
        correction, as the frames of all sequences will be aligned to a common
        reference.

        Parameters
        ----------
        signal_channel : str or int, optional
            Name or index of the dynamic channel to be extracted.
        static_channel : str or int, optional
            Name or index of the static channel. Provide this if you want
            to for example use Suite2p's red cell detection. To determine
            registration shifts using the static channel, override
            'align_by_chan' via ops_kws (see below), setting it equal to 2.
        fs : float, optional
            Frame rate (per-plane), in Hz. Defaults to 10.
        collapse_z : bool, optional
            Whether to average all planes on each frame prior to processing.
            Defaults to True.
        fill_gaps : bool, optional
            Whether to fill NaNed values with data from adjacent frames.
            Defaults to None, which will fill_gaps only if register is False.
        remove_masked_frames : bool, optional
            Skip masked frames when creating binary file. They will be
            reinserted as NaNed entries at the proper timepoints in the final
            imported signal traces.
            Not implemented.
        register : bool, optional
            Whether to use Suite2p motion correction, which is nonrigid by
            default. To use the rigid motion correction only, override
            'nonrigid' via ops_kws (see below), setting it equal to False.
            By default, no registration is performed.
        sparse_mode : bool, optional
            Whether to use Suite2p's sparse mode. Defaults to True
        spatial_scale : int {0, 1, 2, 3}, optional
            ROI scale parameter for sparse model. 0 = multiscale, 1 = 6 pixels,
            2 = 12 pixels, 3 = 24 pixels, 4 = 48 pixels. Choose the closest
            value. Defaults to 0.
        diameter : int or list of ints, optional
            Expected diameter of cells, in pixels. Defaults to None. Ignored
            if sparse_mode is True. If sparse_mode is False, you must specify
            a diameter. Pass a list of two ints to specify different diameters
            for x and y dimensions.
        overwrite : bool, optional
            If True, existing Suite2p folder is deleted and dataset is
            reprocessed from the beginning of the pipeline. Defaults to False.
        reprocess : bool, optional
            If True, rerun Suite2p analysis on existing binary file.
            This can be used for example to extract ROIs multiple times with
            different parameters, without repeating the binary conversion and
            registration each time. Reprocess and overwrite cannot both be
            True.
            Not tested - we may need to rewrite the ops dictionaries.
        ops_kws : dict, optional
            Pass additional keyword-argument pairs as a dictionary to
            override other Suite2p default settings.

        See also
        --------
        lab3.extraction.s2p_helpers.dump_to_binary
        lab3.extraction.s2p_helpers.extract
        """

        # overwrite checks
        assert (not (overwrite & reprocess)), \
            "overwrite and reprocess cannot both be true"
        if os.path.isdir(self.s2p_path):
            if overwrite:
                print('Removing previous Suite2p folder')
                shutil.rmtree(self.s2p_path)
            elif reprocess:
                warnings.warn("reprocess is untested.")
                pass
            else:
                print("Suite2p path:\n%s\n...already exists, exiting..."
                      % self.s2p_path)
                return

        # run extraction
        extract_signals.extract(self, reprocess=reprocess, **kwargs)

    def import_results(self, **kwargs):
        """Import Suite2p signal results using the new-style signal formats
        (dataframes and h5 stores).

        Parameters
        ----------
        ds : Suite2pStrategy
            Instance of a Suite2pStrategy, which has been previously extracted.
        label : str, optional
            ROI label to store the signals and ROI masks under. Defaults to
            'suite2p'.
        channel : {str or int}, optional
            Name or index of dynamic channel in the imaging dataset. Defaults
            to 'Ch2'.
        overwrite : bool, optional
            If key '/{channel}/{label}' already exists in the dataset signals
            file, overwrite with new import. Note this will delete any
            additional signal types in the store in addition to the 'raw' and
            'npil' (e.g. 'dfof'), to prevent mixtures of signals derived from
            different imports. Defaults to False, which will raise an error if
            the desired key already exists.

        See also
        --------
        lab3.extraction.s2p_helpers.import_to_signals_file
        """

        import_signals.import_to_signals_file(self, **kwargs)

    def apply_mc_results(self, ops_list=None, overwrite=False):
        """Apply displacements calculated by Suite2p to the underlying dataset.
        Currently only supports rigid motion correction.

        Parameters
        ----------
        ds : Suite2pStrategy
            Instance of a Suite2pStrategy, which has been previously extracted.
        ops_list : array of dict, optional
            Array containing suite2p options and MC results for each plane.
            Pass only if you want to use suite2p MC results from somewhere else
            If not specified, will look for suite2p results native to this
            dataset.
        overwrite : bool, optional
            Whether to save the motion corrected dataset as a sima directory.
            Will back up the current sima directory, if it exists.
        """

        if ops_list is None:
            ops_list = self.ops_list

        new_sequences = []

        for i in range(self.num_sequences):
            if ops_list[0]['nonrigid']:
                # TODO implement nonrigid sequence wrapper
                raise NotImplementedError
            else:
                seq = _Suite2pRigidSequence(self.sequences[i], ops_list)
                new_sequences.append(seq)

        self.sequences = new_sequences

        if overwrite:
            # backup_dir(self.savedir, delete_original=False)
            self.save()


class OldSuite2pImagingDataset(Suite2pStrategy):

    """For backwards compatibility, this strategy replicates the core
    functionality of the lab's suite2p autoscript, for Suite2p version 0.5.5.
    Additional options are provided to store the extracted signals in the old
    signals format.

    Examples
    --------

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract(self):
        """Not implemented

        See also
        --------
        lab3.extraction.s2p_helpers.extract_with_055
        """

        extract_signals.extract_with_055()

    def import_results(self, as_pickle=False):
        """Not implemented"""

        if as_pickle:
            import_signals.import_to_pkl()
        else:
            raise NotImplementedError



import os

import numpy as np
import pandas as pd
import cv2
from scipy.sparse import csc_matrix
from scipy.optimize import linear_sum_assignment
from skimage.measure import find_contours

from lab3.misc.progressbar import ProgressBar

CV2_CRITERIA = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 1000, 1e-10)


def get_suite2p_time_average(ds):
    """
    Parameters
    ----------
    ds: sima.ImagingDataset

    Returns
    -------
    array (num_planes, num_rows, num_columns)
    """
    ops_list = np.load(os.path.join(ds.savedir, "suite2p", "ops1.npy"),
                       allow_pickle=True)
    return np.stack([ops['meanImgE'] for ops in ops_list])


def crop_image(image, crop):
    """Maybe write a function to do multi-z crops? If only one crop tuple is
    passed, then use it for all planes?"""
    pass


def find_crop_coords(image, threshold=1e-10):
    """Find the columns and rows of an image that are of constant value.

    Parameters
    ----------
    image : array (n_rows, n_columns)
    threshold : float, optional
        Threshold on the variance of values in a row or column, for determining
        whether the data is constant.

    Returns
    -------
    rows, columns : array
        Lists of rows and columns to crop
    """

    rows = image.var(axis=1) > threshold
    columns = image.var(axis=0) > threshold

    return rows, columns


def float_to_uint(image):
    image -= np.min(image)
    image /= np.max(image)
    image *= 255
    return image.astype('uint8')


def compute_transforms(images, reference_idx=0, from_reference=False):
    """Compute an affine transform between each image and a chosen reference
    image.

    Parameters
    ----------
    images : list
    reference_idx : int, optional
        List index of the reference image. Defaults to 0.

    Returns
    -------
    warp_list : list
        List of warp matrices for computed transforms.

    """
    # TODO - add multiprocessing?

    p = ProgressBar(len(images))
    warp_list = []
    for i, current_image in enumerate(images):
        p.update(i)
        warp_list.append(affine_registration(images[reference_idx],
                                             current_image))
    p.update(len(images))
    print(" Done!")
    return warp_list


def affine_registration(image0, image1):
    """Two step registration. First estimate a rigid translation between the
    two images, and then use this to seed the affine transformation
    optimization

    Parameters
    ----------
    image0, image1 : array
        TODO : Which is target, which is reference?
    """

    warp_matrix = np.eye(2, 3, dtype=np.float32)

    image1 = float_to_uint(image1)
    image0 = float_to_uint(image0)

    # First Translate converted raw images
    (_, warp_matrix) = cv2.findTransformECC(image1, image0, warp_matrix,
                                            cv2.MOTION_TRANSLATION,
                                            CV2_CRITERIA, None, 1)

    # Then seed affine transform for processed images
    (_, warp_matrix) = cv2.findTransformECC(image1, image0, warp_matrix,
                                            cv2.MOTION_AFFINE, CV2_CRITERIA,
                                            None, 1)

    return warp_matrix


def calc_distance_matrix(masks0, masks1):
    """Calculates a matrix of Jaccard distances between the columns of the
    binary input matrices, equal to 1 minus the intersection / union. Uses
    sparse matrices to conserve memory when dealing with a large number of
    ROI pairs."""

    flat_masks = np.vstack([np.stack([m.flatten() for m in masks])
                            for masks in [masks0, masks1]])
    mat = csc_matrix(flat_masks.T)

    cols_sum = mat.getnnz(axis=0)
    ab = mat.T * mat
    aa = np.repeat(cols_sum, ab.getnnz(axis=0))
    bb = cols_sum[ab.indices]

    similarities = ab.copy()
    similarities.data = similarities.data / (aa + bb - ab.data)

    return 1 - similarities.todense()[0:len(masks0), len(masks0):]


def binarize_mask(roi, thres=0.05):
    """Return binarized numpy array given an ROI object"""
    mask = np.sum(roi.__array__(), axis=0)
    return mask > (np.max(mask) * thres)


def calc_mask_centroid(mask):
    return [np.mean(l) for l in np.where(mask)]


def calc_mask_contour(mask, level=.99):
    if np.sum(mask) == 0:
        return None
    else:
        return find_contours(mask, level=level)[0].T[::-1]


def apply_transform(image, transform, sz, invert=False):
    if invert:
        transform = cv2.invertAffineTransform(transform)
    return cv2.warpAffine(image.astype(np.uint8), transform, (sz[1], sz[0]),
                          flags=cv2.INTER_LINEAR).astype(np.int)


def get_roi_info(roi, crop=None, transform=None, target_shape=None):
    """Returns some some basic information about an ROI object

    Parameters
    ----------
    roi : sima.ROI
    crop : list of list of arrays
    transform : list of arrays, optional
        Affine transform to apply, per plane
    target_shape : tuple, optional
        (num_planes, num rows, num_columns)

    Returns
    -------
    pd.Series containing fields:
        roi: sima.ROI
        plane: int
        mask: 2D array
        contour: list of arrays

    """
    mask = binarize_mask(roi)
    plane = int(roi.label[3])
    if crop is not None:
        rows, columns = crop[plane]
        mask = mask[rows][:, columns]

    if transform is not None:
        mask = apply_transform(mask, transform[plane], target_shape[1:])

    return pd.Series({'roi': roi,
                      'plane': int(roi.label[3]),
                      'mask': mask,
                      'centroid': calc_mask_centroid(mask),
                      'contour': calc_mask_contour(mask)},
                     name=roi.label)


def find_nearby_rois(centroid, reference_rois, max_distance=25,
                     exclude_partnered=None):
    """Filter the reference ROIs for ones whose centroids are nearby the
    current ROI.

    Parameters
    ----------
    centroid : list
    reference_rois : pd.DataFrame
    max_distance : float, optional
        Maximum centroid distance between the target and reference. Defaults to
        25 pixels
    exclude_partnered : str, optional
        savedir of an ImagingDataset. Exclude any reference ROIs that are
        already partnered to an ROI from this dataset.

    Returns
    -------
    eligible_references : pd.DataFrame
    """
    eligible_references = reference_rois[[exclude_partnered not in partners for
                                          partners in reference_rois.partners]]
    distances = np.abs(np.stack(eligible_references.centroid)
                       - np.asarray(centroid))
    return eligible_references[(distances < max_distance).all(axis=1)]



import os.path
import time
import pickle

import MySQLdb
import MySQLdb.cursors
import json


class ExperimentDatabase:
    """An object to wrap MySQLdb and handle all connections to mySQL server

    Example
    -------

    >>> from lab.classes import database
    >>> db = database.ExperimentDatabase()
    >>> db.select_all('SELECT * FROM mice LIMIT 10')
    >>> db.disconnect()
    """

    _trial_fields = ['trial_id', 'mouse_id', 'start_time', 'stop_time',
                     'behavior_file', 'tSeries_path', 'experiment_group',
                     'experiment_id', 'mouse_name']

    def __init__(self):
        """Connects to SQL database upon initialization."""

        self.connect()

    def connect(self):
        """Connect to the SQL database."""

        host = '127.0.0.1'
        user = 'gergely'       # TODO: use different credentials
        password = 'CipoFuzo1'  # TODO: update credentials
        database = 'experiments'

        self._database = MySQLdb.connect(
            host=host,
            user=user,
            passwd=password,
            db=database,
            cursorclass=MySQLdb.cursors.DictCursor)

    def disconnect(self):
        """Close the connection to SQL database server."""

        self._database.close()

    def select(self, sql, args=[], verbose=False):
        """ Queries SQL database return single result

        Parameters
        ----------
        sql : str
            Raw SQL to pass to the database.
        args : list, optional
            List of variables to sub into SQL statement in accordiance see
            MySQLdb. Defaults to an empty list.
        verbose : bool, optional
            If set to True print the SQL query. Defaults to False

        Returns
        -------
        result : dict
            First record to match SQL query as a dictionary, with the field
            name being the key.
        """

        cursor = self._database.cursor()
        cursor.execute(sql, args)
        if verbose:
            print(sql)

        result = cursor.fetchone()
        return result

    def select_all(self, sql, args=[], verbose=False):
        """ Queries SQL database and return all the results

        Parameters
        ----------
        sql : str
            Raw SQL to pass to the database.
        args : list, optional
            List of variables to sub into SQL statement in accordiance see
            MySQLdb. Defaults to an empty list.
        verbose : bool, optional
            If set to True print the SQL query. Defaults to False

        Returns
        -------
        result : dict
            All records to match SQL query as a dictionary, with the field name
            being the key.
        """

        cursor = self._database.cursor()
        cursor.execute(sql, args)
        if verbose:
            print(sql)

        result = cursor.fetchall()
        return result

    def query(self, sql, args=[], verbose=False, ignore_duplicates=True):
        """ Run an arbitrary SQL query i.e. INSERT or DELETE commands. SQL
        statement is passsed directory MySQLdb.

        Parameters
        ----------
        sql : str
            Raw SQL statement to query the database with.
        args : list, optional
            List of variables to sub into SQL statement in accordiance see
            MySQLdb. Defaults to an empty list.
        verbose : bool, optional
            If set to True print the SQL query. Defaults to False
        ignore_duplicates : bool, optional
            If set to True Duplicate entry errors are prevented from being
            raised, instead of raising an error the method returns False.
            Defaults to True.

        Returns
        -------
        bool : True if the SQL statement executes, False otherwise
        """

        cursor = self._database.cursor()
        try:
            cursor.execute(sql, args)
        except Exception as e:
            if not ignore_duplicates or 'Duplicate entry' not in e.__str__():
                raise e
            return False
        else:
            if verbose:
                print(sql)
            self._database.commit()
        return True


def add_experiment_to_database(filename, trial_id=None, overwrite=False):
    """Add an experiment to the database from data in pickled behavior file.

    Parameters
    ----------
    filename : str
        Path to pickled behavior data for the experiment to be added
    trial_id : int, optional
        Update this trial_id in database. Defaults to None (a new id is created
        if the behavior file is not associated with an existing database entry)
    overwrite : bool, optional
        Whether re-add experiment to the database, if an existing database
        entry is already associated with this behavior data file
    """
    if not overwrite and fetch_trial_id(behavior_file=filename) is not None:
        print(f'trial {filename} already loaded')
        return fetch_trial_id(behavior_file=filename)

    # get trial info from pickled behavior file
    with open(filename, 'rb') as f:
        data = pickle.load(f, encoding='latin1')
    try:
        trial_info = data['__trial_info']
    except KeyError:
        raise Exception(f"__trial_info not found in {filename}")

    update_trial(filename, trial_info['mouse'], trial_info['start_time'],
                 trial_info['stop_time'], trial_info['experiment_group'],
                 trial_id=trial_id)

    if trial_id is None:
        trial_id = fetch_trial_id(filename)
    else:
        delete_all_trial_attrs(trial_id)

    print('loading trial {}: {}'.format(trial_id, filename))

    for key in filter(lambda k: k not in [
            'mouse', 'start_time', 'stop_time', 'experiment_group'],
            trial_info.keys()):

        if key == 'contexts':
            for context in trial_info['contexts'].keys():
                update_trial_attr(trial_id, f'{context}_context',
                                  json.dumps(trial_info[key][context]))
        elif isinstance(trial_info[key], dict) or \
                isinstance(trial_info[key], list):

            # TODO : figure out a more general way to handle long entries?
            if key == 'controllers':
                continue
                # for k, v in trial_info[key].items():
                #     update_trial_attr(trial_id, k, json.dumps(v))
            else:
                update_trial_attr(trial_id, key, json.dumps(trial_info[key]))

        else:
            update_trial_attr(trial_id, key, trial_info[key])

    return trial_id


def fetch_mouse_weights(mouse_name):
    """ Fetch all the weight records for a given mouse in the SQL database.

    Parameters
    ----------
    mouse_name : str
        Name of the mouse to search the database for records of weights.

    Returns
    -------
    weights : list
        List of dictionaries with mouse_name, weight, and date keys.
    """

    db = ExperimentDatabase()
    weights = db.select_all("""
        SELECT mouse_name, weight, date
        FROM mice
        INNER JOIN mouse_weights
        ON mice.mouse_id = mouse_weights.mouse_id
        WHERE mouse_name = %s
        ORDER BY date ASC
        """, args=[mouse_name], verbose=False)

    return weights


def read_mouse_weights(mouse_name, limit=10):
    """ Print out a report of recent mouse weight records and the average
    overall weight for the mouse.

    Parameters
    ----------
    mouse_name : str
        Name of the mouse to search database for weight records of.
    limit : int, optional
        Number of records to print. Defaults to 10.
    """

    db = ExperimentDatabase()
    weights = db.select_all("""
        SELECT mouse_name, weight, date
        FROM mice
        INNER JOIN mouse_weights
        ON mice.mouse_id = mouse_weights.mouse_id
        WHERE mouse_name = %s
        ORDER BY date DESC
        LIMIT %s
        """, args=[mouse_name, limit], verbose=False)

    def print_weight(weight):
        print(f"{weight['date']} \t {weight['weight']}")

    print(f'Recent records for mouse {mouse_name}')
    map(print_weight, weights)

    average = db.select("""
        SELECT AVG(weight)
        FROM mice
        INNER JOIN mouse_weights
        ON mice.mouse_id = mouse_weights.mouse_id
        WHERE mouse_name = %s
        """, args=[mouse_name], verbose=False)
    print(f'\naverage weight: {average.values()[0]}\n\n')


def insert_mouse_weight(mouse_name, weight, date=None):
    """ Insert a new weight record into the mouse weight database. Calls
    read_mouse_weights before completing.

    Parameters
    ----------
    mouse_name : str
        Name of the mouse corresponding to this weight. Matches mouse_name
        in the mice table.
    weight : float
        Weight value to add to the mouse_wieghts table in the database.
    date : str, optional
        Date to associate the weight entry with. Should be in the mySQL date
        format, YYYY-MM-DD HH:MM:SS. Defaults to now.
    """

    db = ExperimentDatabase()

    mouse_ID = int(db.select("""
        SELECT mouse_id
        FROM mice
        WHERE `mouse_name` = %s
        """, args=[mouse_name]).values()[0])

    if date is None:
        db.query("""
            INSERT INTO mouse_weights (`mouse_id`,`weight`)
            VALUES (%s, %s)
            """, [mouse_ID, weight], verbose=False)
    else:
        db.query("""
            INSERT INTO mouse_weights (`mouse_id`,`weight`,`date`)
            SELECT `mouse_id`, %s, %s
            FROM mice
            WHERE `mouse_name` = %s
            """, [weight, date, mouse_name], verbose=False)

    read_mouse_weights(mouse_name)


# def pair_imaging_data(imaging_path, trial_id=None, behavior_file=None,
#                       start_time=None):
#     """ Pair a tSeries directory with a trial record in the database.
#
#     Parameters
#     ----------
#     imaging_path : str
#         tSeries directory to add to trial record.
#     trial_id : int, optional
#         Trial id from the SQL trials table to match SQL record. Defaults to
#         None.
#     behavior_file : str, optional
#         Path to a behavior file which matches a SQL record. Defaults to None.
#     start_time : str, optional
#         Date time string to match a to  SQL record. Defaults to None.
#     """
#
#     db = ExperimentDatabase()
#
#     if trial_id and behavior_file:
#         raise ValueError('Cannot use both trial id and behavior file path')
#     elif trial_id is not None:
#         db.query("""
#             UPDATE trials
#             SET `tSeries_path` = %s
#             WHERE `trial_id` = %s
#             """, args=[imaging_path, trial_id], verbose=False)
#     elif behavior_file is not None:
#         behavior_file = os.path.splitext(behavior_file)[0] + '%'
#         db.query("""
#             UPDATE trials
#             SET `tSeries_path` = %s
#             WHERE behavior_file LIKE %s
#         """, args=[imaging_path, behavior_file], verbose=False)
#         trial_id = fetch_trial_id(behavior_file=behavior_file)
#
#     # TODO: should be start_time/mouse_name combination to ensure this value
#     # is unique.
#     elif start_time is not None:
#         start_time = _resolve_start_time(start_time)
#         # TODO: experiments table is not working (or used) necessary for
#         # legacy saliance compatibilty. Update or remove...
#
#         # expt_id = fetch_experiment_ID(start_time)
#         # if expt_id is not None:
#         #     db.query("""
#         #         UPDATE trials
#         #         SET `tSeries_path` = %s
#         #         WHERE `experiment_id` = %s
#         #     """, args=[imaging_path, expt_id], verbose=False)
#         # else:
#         db.query("""
#             UPDATE trials
#             SET `tSeries_path` = %s
#             WHERE `start_time` = %s
#         """, args=[imaging_path, start_time], verbose=False)
#
#     else:
#         raise ValueError('Need either trial id or behavior file path')
#
#     db.disconnect()


def update_behavior_filename(old_filename, new_filename):
    """ Update the behavior_file field of a SQL record in the trials database.

    Parameters
    ----------
    old_filename : str
        The file path currently stored in the SQL database to look up the
        record by.
    new_filename : str
        The new path to change the behavior_file record to.
    """

    db = ExperimentDatabase()
    db.query("""
        UPDATE trials
        SET `behavior_file` = %s
        WHERE behavior_file = %s
        """, args=[new_filename, old_filename])

    db.disconnect()


def update_trial(behavior_file, mouse_name, start_time, stop_time,
                 experiment_group, trial_id=None):
    """
    Parameters
    ----------
    behavior_file : str
        Path to behavior file.
    mouse_name : str
    start_time : str
    stop_time : str
    experiment_group : str
    trial_id : int, optional
        Target trial to update. If None, a new trial is created."""

    mouse_id = fetch_mouse_ID(mouse_name, create=True)
    if trial_id is None:
        trial_id = fetch_trial_id(behavior_file=behavior_file)
    start_time = _resolve_start_time(start_time)
    stop_time = _resolve_start_time(stop_time)

    db = ExperimentDatabase()
    # if this behavior_file path is alread associated with the trial, update
    # the existing record, otherwise create a new one.
    if trial_id is not None:
        db.query("""
            UPDATE trials
            SET `mouse_id` = %s,
                `start_time` = %s,
                `stop_time` = %s,
                `behavior_file` = %s,
                `experiment_group` = %s
            WHERE trial_id = %s
            """, args=[mouse_id, start_time, stop_time, behavior_file,
                 experiment_group, trial_id])
    else:
        db.query("""
            INSERT INTO trials
                (`mouse_id`, `start_time`, `stop_time`, `behavior_file`,
                 `experiment_group`)
            VALUES (%s, %s, %s, %s, %s)
            """, args=[mouse_id, start_time, stop_time, behavior_file,
                 experiment_group])

    db.disconnect()


def update_attr(table, group_name, group_id, attribute, value):
    """
    Parameters
    ----------
    table :
    group_name :
    group_id :
    attribute : str
        Attribute in database to set
    value : str
        Value to set for attribute
    """

    db = ExperimentDatabase()
    attr = db.select("""
        SELECT attribute_id
        FROM {}
        WHERE {} = %s
        AND attribute = %s
        """.format(table, group_name), args=[group_id, attribute],
        verbose=False)

    if attr is not None:
        db.query("""
            UPDATE {}
            SET value = %s
            WHERE attribute_id = %s
            """.format(table),
                 args=[value, attr['attribute_id']], verbose=False)
    else:
        db.query("""
            INSERT INTO {} (`{}`, `attribute`, `value`)
            VALUES (%s, %s, %s)
            """.format(table, group_name),
                 args=[group_id, attribute, value], verbose=False)

    db.disconnect()


def fetch_attr(table, group_name, group_id, attribute, default=None):
    """
    Parameters
    ----------
    table :
    group_name :
    group_id :
    attribute : str
        Attribute to fetch
    default : optional
        What to return if attribute does not exist.
    """
    db = ExperimentDatabase()
    attr = db.select("""
        SELECT value
        FROM {}
        WHERE {} = %s
        AND attribute = %s
        """.format(table, group_name),
        args=[group_id, attribute], verbose=False)
    db.disconnect()

    if attr:
        return attr['value']

    return default


def _delete_all_attrs(table, group_name, group_id):
    db = ExperimentDatabase()
    db.query("""
        DELETE FROM {}
        WHERE {} = %s
        """.format(table, group_name), args=[group_id], verbose=False)
    db.disconnect()


def _delete_attr(table, group_name, group_id, attribute):
    db = ExperimentDatabase()
    db.query("""
        DELETE FROM {}
        WHERE {} = %s
        AND attribute = %s
        """.format(table, group_name),
             args=[group_id, attribute], verbose=False)
    db.disconnect()


def update_mouse_attr(mouse_name, attribute, value):
    mouse_id = fetch_mouse_ID(mouse_name)
    update_attr('mouse_attributes', 'mouse_id', mouse_id, attribute, value)


def update_mouse_page(mouse_name, attribute, value):
    mouse_id = fetch_mouse_ID(mouse_name)
    if type(value) != str:
        value = json.dumps(value)

    update_attr('mouse_pages', 'mouse_id', mouse_id, attribute, value)


def fetch_mouse_attr(mouse_name, attribute, default=None):
    mouse_id = fetch_mouse_ID(mouse_name)
    return fetch_attr(
        'mouse_attributes', 'mouse_id', mouse_id, attribute, default=default)


def delete_mouse_attr(mouse_name, attribute, default=None):
    mouse_id = fetch_mouse_ID(mouse_name)
    _delete_attr(
        'mouse_attributes', 'mouse_id', mouse_id, attribute)


def fetch_mouse_page_attr(mouse_name, attribute, default=None):
    mouse_id = fetch_mouse_ID(mouse_name)
    attr = fetch_attr(
        'mouse_pages', 'mouse_id', mouse_id, attribute, default=default)
    if type(attr) == str:
        try:
            attr = json.loads(attr)
        except ValueError:
            pass

    return attr


def delete_mouse_page_attr(mouse_name, attribute):
    mouse_id = fetch_mouse_ID(mouse_name)
    _delete_attr(
        'mouse_pages', 'mouse_id', mouse_id, attribute)


def update_trial_attr(trial_id, attribute, value):
    update_attr('trial_attributes', 'trial_id', trial_id, attribute, value)


def fetch_trial_attr(trial_id, attribute, default=None):
    return fetch_attr(
        'trial_attributes', 'trial_id', trial_id, attribute, default=default)


def delete_trial_attr(trial_id, attribute):
    _delete_attr(
        'trial_attributes', 'trial_id', trial_id, attribute)


def update_virus_attr(trial_id, attribute, value):
    update_attr('virus_attributes', 'virus_id', trial_id, attribute, value)


def fetch_virus_attr(trial_id, attribute, default=None):
    return fetch_attr(
        'virus_attributes', 'virus_id', trial_id, attribute, default=default)


def delete_virus_attr(trial_id, attribute):
    _delete_attr(
        'virus_attributes', 'virus_id', trial_id, attribute)


def _fetch_all_attrs(table, id_field, id_value, parse=False):
    db = ExperimentDatabase()
    attrs = db.select_all("""
        SELECT attribute, value
        FROM {}
        WHERE {} = %s
        """.format(table, id_field), args=[id_value], verbose=False)
    db.disconnect()

    attrs = {entry['attribute']: entry['value'] for entry in attrs}
    if parse:
        for key in attrs.keys():
            try:
                attrs[key] = json.loads(attrs[key])
            except ValueError:
                pass

    return attrs


def fetch_all_virus_attrs(virus_id, parse=True):
    attrs = _fetch_all_attrs('virus_attributes', 'virus_id', virus_id)

    return attrs


def fetch_all_mouse_attrs(mouse_name, parse=False):
    if isinstance(mouse_name, int):
        mouse_id = mouse_name
    else:
        mouse_id = fetch_mouse_ID(mouse_name)

    return _fetch_all_attrs('mouse_attributes', 'mouse_id', mouse_id,
                            parse=parse)


def fetch_all_trial_attrs(trial_id, parse=False):
    db = ExperimentDatabase()
    attrs = db.select_all("""
        SELECT attribute, value
        FROM trial_attributes
        WHERE trial_id = %s
        """, args=[trial_id], verbose=False)
    db.disconnect()

    if parse:
        for i, entry in enumerate(attrs):
            try:
                attrs[i]['value'] = json.loads(entry['value'])
            except ValueError:
                pass

    return {entry['attribute']: entry['value'] for entry in attrs}


def delete_all_trial_attrs(trial_id):
    _delete_all_attrs('trial_attributes', 'trial_id', trial_id)


def delete_all_mouse_attrs(mouse_id):
    _delete_all_attrs('mouse_attributes', 'mouse_id', mouse_id)
    _delete_all_attrs('mouse_pages', 'mouse_id', mouse_id)


def _project_filter_sql(project_name):
    return """
        (SELECT DISTINCT m.*
         FROM mice m
         LEFT JOIN trials t
            ON m.mouse_id=t.mouse_id
         WHERE experiment_group='{0}'
         UNION
         SELECT DISTINCT m.*
         FROM mice m
         LEFT JOIN mouse_attributes ma
            ON m.mouse_id=ma.mouse_id
         WHERE (attribute='project_name' AND value='{0}')
         )
    """.format(project_name)


def delete_trial(trial_id):
    delete_all_trial_attrs(trial_id)

    db = ExperimentDatabase()
    db.query("""
        DELETE FROM trials
        WHERE trial_id = %s
        """, args=[trial_id])


def delete_mouse(mouse_id):
    delete_all_mouse_attrs(mouse_id)

    db = ExperimentDatabase()
    db.query("""
        DELETE FROM mice
        WHERE mouse_id = %s
        """, args=[mouse_id], verbose=False)


def fetch_mouse_ID(mouse_name, create=False, project_name=None):
    db = ExperimentDatabase()

    if project_name is not None:
        mouse_id = db.select_all("""
            SELECT DISTINCT mice.mouse_id
            FROM mice
            INNER JOIN trials
                ON trials.mouse_id=mice.mouse_id
            WHERE mouse_name = %s
                AND experiment_group = %s
            UNION
            SELECT mice.mouse_id
            FROM mice
            INNER JOIN mouse_attributes
                ON mice.mouse_id = mouse_attributes.mouse_id
            WHERE mouse_name = %s
                AND attribute = %s
                AND value = %s
            """, args=[mouse_name, project_name, mouse_name, 'project_name',
                       project_name])

    else:
        mouse_id = db.select_all("""
            SELECT mouse_id
            FROM mice
            WHERE mouse_name = %s
            """, args=[mouse_name])

    if create and len(mouse_id) == 0:
        db.query("""
            INSERT INTO mice (mouse_name)
            VALUES (%s)
        """, args=[mouse_name])
        db.disconnect()
        if project_name is not None:
            update_mouse_attr(mouse_name, 'project_name', project_name)
        mouse_id = fetch_mouse_ID(mouse_name, create=False)
    elif len(mouse_id) == 1:
        db.disconnect()
        mouse_id = int(list(mouse_id[0].values())[0])
    else:
        raise KeyError('unable to uniquely identify mouse {}'.format(
            mouse_name))

    return mouse_id


def fetch_mouse(mouse_id):
    db = ExperimentDatabase()
    mouse = db.select("""
        SELECT *
        FROM mice
        WHERE mouse_id = %s
    """, args=[mouse_id])

    return mouse


def fetch_trial(trial_id):
    db = ExperimentDatabase()
    trial = db.select("""
        SELECT *
        FROM trials
        WHERE trial_id = %s
        """, args=[trial_id])
    return trial


def fetch_trial_id(behavior_file=None, tSeries_path=None, mouse_name=None,
                   startTime=None):
    db = ExperimentDatabase()

    arg_names = ['tSeries_path', 'behavior_file', 'mouse_name', 'startTime']
    arg_vals = [tSeries_path, behavior_file, mouse_name, startTime]
    arg = [n for n, v in zip(arg_names, arg_vals) if v is not None]

    if 'startTime' in arg:
        trial_id = db.select("""
            SELECT trial_id FROM trials
            INNER JOIN mice
                ON trials.mouse_id=mice.mouse_id
            WHERE mouse_name = %s
                AND start_time = %s
        """, args=[mouse_name, _resolve_start_time(startTime)])
    elif 'startTime' in arg:
        raise Exception("require mouse_id")
    else:
        trial_id = db.select("""
            SELECT trial_id FROM trials
            WHERE {} = %s
            """.format(arg[0]), args=[eval(arg[0])])

    if trial_id is None:
        return None

    return int(trial_id['trial_id'])


def fetch_imaged_trials(mouse_name):
    db = ExperimentDatabase()
    trials = db.select_all("""
        SELECT trial_id, start_time, mouse_name, behavior_file, tSeries_path
        FROM trials
        INNER JOIN mice
        ON mice.mouse_id = trials.mouse_id
        WHERE tSeries_path IS NOT NULL
        AND mouse_name = %s
        ORDER BY start_time ASC
        """, args=[mouse_name])
    db.disconnect()

    return trials


def fetch_attribute_values(attr, project_name=None):
    db = ExperimentDatabase()

    table = 'mice'
    if attr in db._trial_fields:
        condition = ''
        if project_name is not None:
            condition = 'WHERE experiment_group=\'{}\''.format(project_name)

        values = db.select_all("""
            SELECT DISTINCT {} FROM trials t
            LEFT JOIN mice m
                ON t.mouse_id=m.mouse_id
            {}
        """.format(attr, condition))
        return [value[attr] for value in values]

    if project_name is not None:
        table = _project_filter_sql(project_name)

    values = db.select_all("""
        SELECT DISTINCT value FROM {0} AS m
        LEFT JOIN trials t
            ON m.mouse_id=t.mouse_id
        LEFT JOIN trial_attributes ta
            ON t.trial_id=ta.trial_id
        WHERE attribute=%s
        UNION
        SELECT DISTINCT value FROM {0} AS m
        LEFT JOIN mouse_attributes ma
            ON m.mouse_id=ma.mouse_id
        WHERE attribute=%s
    """.format(table), args=[attr, attr])

    return [value['value'] for value in values]


def fetch_trials_with_attr_value(attr, value):
    db = ExperimentDatabase()
    trials = db.select_all("""
        SELECT trial_id
        FROM trial_attributes
        WHERE attribute = %s AND value = %s
        ORDER BY trial_id ASC
        """, args=[attr, value])
    db.disconnect()

    return [int(trial['trial_id']) for trial in trials]


def fetch_mice(*args, **kwargs):
    db = ExperimentDatabase()
    query_string = """
        SELECT DISTINCT m.mouse_id
        FROM {0} AS m
        LEFT JOIN mouse_attributes ma
            ON m.mouse_id=ma.mouse_id
        LEFT JOIN trials t
            ON m.mouse_id=t.mouse_id
        WHERE {1}
        UNION
        SELECT DISTINCT m.mouse_id
        FROM {0} AS m
        LEFT JOIN trials t
            ON m.mouse_id=t.mouse_id
        LEFT JOIN trial_attributes ta
            ON t.trial_id=ta.trial_id
        WHERE {1}
        """

    table = 'mice'
    if 'project_name' in kwargs.keys():
        table = _project_filter_sql(kwargs['project_name'])
        del kwargs['project_name']

    if 'experiment_group' in kwargs.keys():
        table = _project_filter_sql(kwargs['experiment_group'])
        del kwargs['experiment_group']

    trial_conditions = []
    conditions = []
    query_args = []
    trial_condition = "({} IS NOT NULL)"
    condition_string = "(attribute='{}')"
    for key in args:
        if key in db._trial_fields:
            if key == 'mouse_id':
                key = 'm.%s' % key

            trial_conditions.append(trial_condition.format(key))
        else:
            conditions.append(condition_string.format(key))

    trial_condition = "({}={})"
    condition_string = "(attribute='{}' AND value={})"
    for key, values in kwargs.items():
        if type(values) != list:
            values = [values]

        alternatives = []
        trial_cond = False
        for val in values:
            if key in ['start_time', 'stop_time']:
                val = _resolve_start_time(val)

            try:
                float(val)
            except BaseException:
                val = "'{}'".format(val)

            if key in db._trial_fields:
                trial_cond = True
                if key == 'mouse_id' or key == 'trial_id':
                    _key = 't.%s' % key
                else:
                    _key = key

                if _key == 'tSeries_path':
                    alternatives.append('tSeries_path LIKE %s')
                    query_args.append(os.path.normpath(val[1:-1] + '%'))
                else:
                    alternatives.append(trial_condition.format(_key, val))
            else:
                alternatives.append(condition_string.format(key, val))
        if trial_cond:
            trial_conditions.append("({})".format(" OR ".join(alternatives)))
            trial_cond = False
        else:
            conditions.append("({})".format(" OR ".join(alternatives)))

    query = None
    if len(trial_conditions):
        query = """
            SELECT DISTINCT m.*
            FROM {} m
            INNER JOIN trials t
                ON m.mouse_id=t.mouse_id
            WHERE {}
        """.format(table, " AND ".join(trial_conditions))

    for condition in conditions:
        if query is None:
            query = query_string.format(table, condition)
        else:
            query = """
                SELECT DISTINCT m.*
                FROM ({}) AS m
                INNER JOIN ({}) as n
                ON m.mouse_id=n.mouse_id
                """.format(query, query_string.format('mice', condition))

    if query is None:
        query = """
            SELECT DISTINCT m.*
            FROM {} AS m
        """.format(table)

    records = db.select_all(query, args=query_args)
    return [int(record['mouse_id']) for record in records]


def fetch_trials(*args, **kwargs):
    db = ExperimentDatabase()

    query_string = """
        SELECT DISTINCT t.*
        FROM {}
        LEFT JOIN trial_attributes ta
            ON t.trial_id = ta.trial_id
        LEFT JOIN mice m
            ON t.mouse_id = m.mouse_id
        WHERE {} ORDER BY start_time"""

    conditions = []
    query_args = []
    trial_condition = "({} IS NOT NULL)"
    condition_string = "(attribute='{}')"
    for key in args:
        if key in db._trial_fields:
            if key == 'mouse_id':
                key = 'trials.%s' % key
            conditions.append(trial_condition.format(key))
        else:
            conditions.append(condition_string.format(key))

    trial_condition = "({}={})"
    condition_string = "(attribute='{}' AND value={})"
    for key, values in kwargs.items():
        if type(values) != list:
            values = [values]

        alternatives = []
        for val in values:
            if key in ['start_time', 'stop_time']:
                val = _resolve_start_time(val)

            try:
                float(val)
            except BaseException:
                val = "'{}'".format(val)

            if key in db._trial_fields:
                if key == 'mouse_id' or key == 'trial_id':
                    _key = 't.%s' % key
                else:
                    _key = key

                if _key == 'tSeries_path':
                    if val[1:-1].strip() == '':
                        alternatives.append('tSeries_path IS NOT NULL')
                    else:
                        alternatives.append('tSeries_path LIKE %s')
                        query_args.append(os.path.normpath(val[1:-1]) + '%')
                else:
                    alternatives.append(trial_condition.format(_key, val))
            else:
                alternatives.append(condition_string.format(key, val))
        conditions.append("({})".format(" OR ".join(alternatives)))

    query = 'trials t'
    for condition in conditions[::-1]:
        query = " (" + query_string.format(query, condition) + ") AS t"
    query = query.rstrip(" AS t")

    trials = db.select_all(query, args=query_args)
    return [int(trial['trial_id']) for trial in trials]


def fetch_mouse_trials(mouse_name):
    db = ExperimentDatabase()
    trials = db.select_all("""
        SELECT trial_id, start_time, mouse_name, behavior_file, tSeries_path
        FROM trials
        INNER JOIN mice
        ON mice.mouse_id = trials.mouse_id
        AND mouse_name = %s
        ORDER BY start_time ASC
        """, args=[mouse_name])
    db.disconnect()

    return trials


def _resolve_start_time(start_time):
    try:
        tstruct = time.strptime(start_time, '%Y-%m-%d %H:%M:%S')
    except ValueError:
        pass
    else:
        return start_time

    formats = ['%Y-%m-%d-%Hh%Mm%Ss', '%m/%d/%Y']
    for t_format in formats:
        try:
            tstruct = time.strptime(start_time, t_format)
            start_time = time.strftime('%Y-%m-%d %H:%M:%S', tstruct)
        except BaseException:
            pass
        else:
            return start_time

    raise Exception('unable to parse time')


def fetch_experiment_ID(start_time):
    db = ExperimentDatabase()
    start_time = _resolve_start_time(start_time)

    experiment_id = db.select("""
        SELECT experiment_id
        FROM experiments
        WHERE start_time = %s
        """, args=[start_time])

    if experiment_id is None:
        experiment_id = db.select("""
            SELECT trial_id
            FROM trials
            WHERE start_time = %s
            """, args=[start_time])

        if experiment_id is None:
            return None

        return int(experiment_id.values()[0])

    return int(experiment_id.values()[0])


def fetch_all_projects():
    db = ExperimentDatabase()
    projects = [record['experiment_group'] for record in
                db.select_all("""
                    SELECT DISTINCT experiment_group
                    FROM trials
                    ORDER BY experiment_group ASC
                """)]
    db.disconnect()

    return projects


def fetch_all_mice(project_name=None):
    db = ExperimentDatabase()

    if project_name is not None:
        mice = [record['mouse_name'] for record in
                db.select_all("""
                    SELECT DISTINCT m.mouse_name
                    FROM mice m
                    LEFT JOIN trials t
                       ON m.mouse_id=t.mouse_id
                    WHERE experiment_group='{0}'
                    UNION
                    SELECT DISTINCT m.mouse_name
                    FROM mice m
                    LEFT JOIN mouse_attributes ma
                       ON m.mouse_id=ma.mouse_id
                    WHERE (attribute='project_name' AND value='{0}')
                    ORDER BY mouse_name
                    """.format(project_name))]
    else:
        mice = [record['mouse_name'] for record in
                db.select_all("""
                    SELECT DISTINCT mouse_name
                    FROM mice
                    ORDER BY mouse_name ASC
                """)]

    return mice


def insert_into_experiment(trial_id, start_time, stop_time=None):
    db = ExperimentDatabase()
    start_time = _resolve_start_time(start_time)

    experiment_id = fetch_experiment_ID(start_time)
    if experiment_id is None:
        db.query("""
            INSERT INTO experiments (start_time)
            VALUES (%s)
            """, args=[start_time])

        experiment_id = fetch_experiment_ID(start_time)

    db.query("""
        UPDATE trials
        SET experiment_id = %s
        WHERE trial_id = %s
        """, args=[experiment_id, trial_id])

    if stop_time is not None:
        db.query("""
            UPDATE experiments
            SET stop_time = %s
            WHERE experiment_id = %s
            """, args=[experiment_id, trial_id])

    db.disconnect()


def fetchVirus(virus_id):
    db = ExperimentDatabase()
    virus = db.select("""
        SELECT *
        FROM viruses
        WHERE virus_id = %s
        """, args=[virus_id])
    db.disconnect()
    return virus


def update_virus(virus_id, name=None, arrival_date=None, source_code=None):
    db = ExperimentDatabase()
    if arrival_date is not None:
        arrival_date = _resolve_start_time(arrival_date)

    id_check = db.select("""
        SELECT COUNT(virus_id)
        FROM viruses
        WHERE virus_id = %s
        """, args=[virus_id])

    if id_check.values()[0] == 0:
        db.query("""
            INSERT INTO viruses
            VALUES (%s, %s, %s, %s)
            """, args=[virus_id, name, arrival_date, source_code])
    else:
        args = ['name', 'arrival_date', 'source_code']
        arg = ['{} = %s'.format(a) for a in args if eval(a) is not None]
        vals = [eval(a) for a in args if eval(a) is not None]
        db.query("""
            UPDATE viruses
            SET {}
            WHERE virus_id = %s
            """.format(', '.join(arg)), args=(vals + [virus_id]))

    db.disconnect()



"""Methods for handling EEG data -- GT 
"""

import pandas as pd

def upsample(df, freq):
    """
    Upsampling dataframes, alternative for pd.resample method which results in less rows.
    If you resample you don't change the boundaries. You still have 4 initial values, 
    and you add as many values as there are intermediates (4 values -> 3 intermediates, 
    n values -> n-1 intermediates)

    Parameters:
    ===========
    df: pandas DataFrame with a Timedelta index or timedelta column
    freq: str
        the upsampling frequecy. E.g. '0.5S', '0.05S'

    Returns:
    ========
    upsampled dataframe

    Note: the `closed` arguent has changed to `inclusive` in pandas >=1.4
	"""

    delta = df.index[1]-df.index[0]
    idx = pd.date_range(df.index.min(),
                        df.index.max()+delta,
                        closed='left',
                        freq=freq)

    return df.reindex(idx).ffill()

def xcorr_comparision(im_exp, df1, df2, st1, st2, save=False):
    """
    Compares two cross correlation matrices.

    Parameters:
    ===========
    im_exp: lab3 ImagigExperiment object
    df1, df2: pandas DataFrames to compare
    st1, st2: str
        brain states belong to the dataframes
    save: boolean, optional

    Return:
    =======
    fig: figure object"""

    state1 = upper(df1.corr())
    state2 = upper(df2.corr())
    ecdf_state1= ECDF(state1)
    ecdf_state2 = ECDF(state2)

    fig, axs = plt.subplots(ncols=2, nrows=2, figsize=(10,8))

    corr_plot(df1, title=st1, ax=axs[0][0],
                xticklabels=False, yticklabels=False, cbar=False)
    corr_plot(df2, title=st2, ax=axs[0][1],
                xticklabels=False, yticklabels=False)

    axs[1][0].plot(ecdf_state1.x, ecdf_state1.y, label=st1)
    axs[1][0].plot(ecdf_state2.x, ecdf_state2.y, label=st2)
    axs[1][0].spines['top'].set_visible(False)
    axs[1][0].spines['right'].set_visible(False)
    axs[1][0].legend()
    axs[1][0].set_title('Cumulative Distribution of pearson r-s')

    significance_test(df1.corr(), df2.corr(), ax=axs[1][1])
    fig.suptitle('Ca activity cross correlation of significanlty upregulated cells')

    if save:
        fig.savefig(join(im_exp.sima_path,
                            f'Ca_xcorr_w_CDF_{st1}_vs_{st2}_states.png'),
                dpi=300, transparent=False, tight_layout=True)
    return fig


"""Temporary dump of functions from Zhenrui's imaging/behavior data pairing
script, to fold it into the functionality of the new ExperimentGroup objects.

TODO : clean up and remove some of the redundancy. Put the remaining
necessary functions in experiment.utils?
"""

import os
import dateutil
from xml.etree import ElementTree as ET

import pandas as pd
import numpy as np
import scipy.optimize as opt

from lab3 import BehaviorExperiment


def get_xmls(toplevel, mouse_name='', keywords=()):
    xmls = []
    for dirpath, dirnames, filenames in os.walk(toplevel):
        if mouse_name in dirpath:
            for f in filenames:
                keywords_in_f = True
                for kwd in keywords:
                    keywords_in_f *= kwd in f
                if keywords_in_f and '.xml' in f and f[0] != '.':
                    xmls.append(os.path.join(dirpath, f))
    return xmls


def xml2sima(xml):
    sima_path = os.path.splitext(xml)[0] + '.sima'
    if os.path.isdir(sima_path):
        return sima_path
    else:
        return np.nan


def get_xml_timestamp(xml):
    try:
        tree = ET.parse(xml)
    except Exception:
        from lxml import etree
        parser = etree.XMLParser(recover=True)
        tree = etree.parse(xml, parser=parser)
    root = tree.getroot()
    ts = dateutil.parser.parse(root.get('date'))
    return ts


def build_xml_series(toplevel, mouse_name='', keywords=()):
    timestamp_df = pd.DataFrame(get_xmls(toplevel, mouse_name=mouse_name,
                                         keywords=keywords),
                                columns=['xml_path'])
    timestamp_df['sima_path'] = timestamp_df['xml_path'].map(xml2sima)
    timestamp_df['timestamp'] = timestamp_df['xml_path'].map(get_xml_timestamp)
    del timestamp_df['xml_path']
    timestamp_df = timestamp_df.dropna()
    return timestamp_df.set_index('sima_path')


def build_expt_series(mouse):
    """
    Parameters
    ----------
    mouse : lab3.experiment.group.Mouse
    """
    return pd.DataFrame([{'trial_id': expt.trial_id,
                          'timestamp': expt.start_time}
                         for expt in mouse]).set_index('trial_id')


def find_pairs(xml_ts, expt_ts, max_dt=100):
    def abs_time_difference(x):
        return np.abs((xml_ts.loc[x[0]]
                       - expt_ts.loc[x[1]]).timestamp.total_seconds())

    product = pd.MultiIndex.from_product([xml_ts.index, expt_ts.index])
    diffs = pd.Series(product.map(abs_time_difference), index=product)
    diffs = diffs[diffs < max_dt]
    cost_mtx = diffs.unstack()
    cost_mtx = cost_mtx.fillna(100000)
    pairs = opt.linear_sum_assignment(cost_mtx)
    square_df = cost_mtx.iloc[pairs]
    paired_expts = pd.DataFrame(np.diag(square_df),
                                index=[square_df.index, square_df.columns],
                                columns=['tdiff'])
    paired_expts = paired_expts[paired_expts['tdiff'] <= max_dt]
    print(f"Found {len(paired_expts)} pairs from {len(xml_ts)} simas and "
          + f"{len(expt_ts)} behavior experiments")

    unpaired_expts = set(xml_ts.index) \
        - set(paired_expts.index.get_level_values('sima_path'))
    for sima_path in unpaired_expts:
        print(f"FAILED TO PAIR: {sima_path}")

    return paired_expts.reset_index()


def pair_database(pair, xml_ts, expt_ts, verbose=False, do_nothing=False,
                  **kwargs):
    extra_info = ""
    if verbose:
        extra_info += f"(IMAGING: {xml_ts.loc[pair.sima_path].timestamp}"
        extra_info += ", "
        extra_info += f"BEHAVIOR: {expt_ts.loc[pair.trial_id].timestamp})"

    print(f"Pairing Trial: {pair.trial_id} with {pair.sima_path} "
          + f"(DELTA={pair.tdiff} s) {extra_info}")
    if not do_nothing:
        expt = BehaviorExperiment(pair.trial_id)
        expt.pair_imaging_data(pair.sima_path, **kwargs)


def pair_mouse_experiments(mouse, imaging_path, max_dt=60, force_pairing=False,
                           keywords=(), do_nothing=False, verbose=False):
    """
    Parameters
    ----------
    mouse : lab3.experiment.group.Mouse
        Mouse to process
    imaging_path : str
        Locate Prairie xmls/sima folders below this folder
    max_dt : float, optional
        Maximum timestamp difference (in seconds) between behavior tdml and
        Prairie xml files. Defaults to 60
    force_pairing : bool, optional
        Whether to overwrite existing sima pairs. Defaults to False
    keywords : iterable of strs, optional
        Additional keywords to identify in sima_paths
    do_nothing : bool, optional
        Find pairs but don't update the database. Defaults to False
    verbose : bool, optional
        Display extra information. Defaults to False
    """

    print(f"Getting behavior timestamps for {mouse.name}...")
    expt_ts = build_expt_series(mouse)

    print(f"Getting imaging xmls for {mouse.name}...")
    xml_ts = build_xml_series(imaging_path, mouse_name=mouse.name,
                              keywords=keywords)

    print("Pairing using Hungarian algorithm...")
    pairs = find_pairs(xml_ts, expt_ts, max_dt=max_dt)
    pairs.apply(pair_database, axis=1, xml_ts=xml_ts, expt_ts=expt_ts,
                verbose=verbose, do_nothing=do_nothing,
                force_pairing=force_pairing)



import os
import shutil
import datetime


def backup_dir(directory, delete_original=False):
    directory = directory.rstrip(os.sep)
    backup = '{}.backup_{}'.format(
        directory, datetime.datetime.today().strftime('%Y-%m-%dT%H-%M-%S'))

    if delete_original:
        shutil.move(directory, backup)
    else:
        shutil.copytree(directory, backup)


def backup_sima(imds):
    sima_dir = imds.savedir
    backup_dir(sima_dir)


def get_base_sequence(seq):
    """Strips away any existing sequence wrappers to access the raw data"""
    base = seq
    while True:
        try:
            base = base._base
        except AttributeError:
            return base



import sys

import numpy as np
import scipy.stats as sp
from scipy.special import factorial
from sklearn.cluster import KMeans
from sklearn.utils import check_random_state

try:
    from hmmlearn.base import _BaseHMM, ConvergenceMonitor
    from hmmlearn.utils import iter_from_X_lengths

except ImportError:
    raise ImportError("hmmlearn is not installed")

np.seterr(divide='ignore')


class RelativeConvergenceMonitor(ConvergenceMonitor):

    """A modified convergence monitor with more relaxed stopping criteria.
    Rather than evaluate the tolerance on the difference of likelihoods at
    iterations t and t_-1, we use the percentage change, i.e. the ratio of
    (t - t_-1) / t_-1."""

    # what do we need to override?

    def report(self, logprob):
        """Reports convergence to :data:`sys.stderr`.
        The output consists of three columns: iteration number, log
        probability of the data at the current iteration and convergence
        rate.  At the first iteration convergence rate is unknown and
        is thus denoted by NaN.

        Parameters
        ----------
        logprob : float
            The log probability of the data as computed by EM algorithm
            in the current iteration.
        """
        if self.verbose:
            delta = logprob - self.history[-1] if self.history else np.nan
            relative_delta = np.abs(delta / self.history[-1]) \
                if self.history else np.nan

            message = self._template.format(
                iter=self.iter + 1, logprob=logprob, delta=relative_delta)
            print(message, file=sys.stderr)

        self.history.append(logprob)
        self.iter += 1

    @property
    def converged(self):
        """``True`` if the EM algorithm converged and ``False`` otherwise."""
        if self.iter == self.n_iter:
            return True
        else:
            if len(self.history) == 2:
                delta = self.history[1] - self.history[0]
                relative_delta = np.abs(delta / self.history[1])
                return relative_delta < self.tol
            else:
                return False


class BaseMixin:

    """Mixin to fix issue in hmmlearn's _BaseHMM, which uses a deprecated
    private methods from sklearn's BaseEstimator class."""

    def _init(self, X, lengths):
        """Initializes model parameters prior to fitting.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix of individual samples.
        lengths : array-like of integers, shape (n_sequences, )
            Lengths of the individual sequences in ``X``. The sum of
            these should be ``n_samples``.
        """
        init = 1. / self.n_components
        if 's' in self.init_params or not hasattr(self, "startprob_"):
            self.startprob_ = np.full(self.n_components, init)
        if 't' in self.init_params or not hasattr(self, "transmat_"):
            self.transmat_ = np.full((self.n_components, self.n_components),
                                     init)


class ScoringMixin:

    """Mixin class that provides some auxilary methods for scoring the models.
    Useful for doing model selection"""

    def bic(self, X, lengths=None):
        """Bayesian information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        lengths : array of shape (n_trials,)
            Note sum(lengths)==n_samples

        Returns
        -------
        bic : float
            The lower the better.
        """
        return -2 * self.score(X, lengths=lengths) \
            + (self.n_components * (self.n_components - 1)
               + self.n_components * self.n_features) * np.log(X.shape[0])

    def aic(self, X, lengths=None):
        """Akaike information criterion for the current model on the input X.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        lengths : array of shape (n_trials,)
            Note sum(lengths)==n_samples

        Returns
        -------
        aic : float
            The lower the better.
        """
        return -2 * self.score(X, lengths=lengths) \
            + (self.n_components * (self.n_components - 1)
               + self.n_components * self.n_features)

    def estimate_transitions(self, X, lengths=None, method='empirical',
                             algorithm='viterbi'):
        """Empirically estimate the transition matrix from the observation
        sequence(s). State transitions are tabulated along the Viterbi path
        by default.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        lengths : array of shape (n_sequences), optional
            Length of each observation sequence in X.
            Note sum(lengths)==n_samples

        algorithm : str, optional
            Either "viterbi" or "map"

        Returns
        -------
        trans_mat_list : array of shape
                         (n_sequences, n_components, n_components)
            Empirical transition matrix for each observation sequence.

        # TODO could do this analytically as well
        """

        if lengths is None:
            lengths = [X.shape[0]]

        state_path = self.decode(X, lengths=lengths, algorithm=algorithm)[1]

        # cum_length = 0
        trans_mat_list = []
        # for length in lengths:
        for i, j in iter_from_X_lengths(X, lengths):
            if method == 'empirical':
                path_seg = state_path[i:j]

                trans_mat = np.zeros((self.n_components, self.n_components))
                for t in range(len(path_seg[:-1])):
                    trans_mat[path_seg[t], path_seg[t+1]] += 1

                trans_mat_list += [(trans_mat.T / np.sum(trans_mat, axis=1)).T]
            elif method == 'analytical':
                # TODO
                # Compute the m-step estimates of the transitions for each
                # trial
                raise NotImplementedError

        return np.stack(trans_mat_list)

    def estimate_emissions(self, X, lengths=None, dt=0.125,
                           method='empirical', algorithm='viterbi'):
        """Estimate the emission matrix from the observation sequence(s).
        Emissions can be estimated analytically from the Baum-Welch m-step,
        or empirically from the inferred state sequence.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        lengths : array of shape (n_sequences), optional
            Length of each observation sequence in X.
            Note sum(lengths)==n_samples

        dt : float, optional
            Length of sample bin in X in seconds, default is 0.125. Estimated
            emissions are converted to firing rates via emissions * 1/dt.

        method : str, optional
            'empirical' (default) or 'analytical'

        algorithm : str, optional
            'viterbi' (default) or 'map'

        Returns
        -------
        emissions_list : array of shape(n_sequences, n_components, n_features)
            Estimated emissions matrix for each observation sequence
        """

        # emissions_list = []
        # if method=='analytical':
        #     posterior = self.predict_proba(X, lengths=lengths)
        #
        #     emissions_list = []
        #     for i, j in iter_from_X_lengths(X, lengths):
        #         num = np.dot(X[i:j].T, posterior[i:j])
        #         den = np.sum(posterior[i:j], axis=0)
        #
        #         emissions_list += [-1 * (1/dt) * np.log(1 - (num/den))]
        #
        # elif method=='empirical':
        #     for i, j in iter_from_X_lengths(X, lengths):
        #         state_path = self.decode(X[i:j], algorithm=algorithm)[1]
        #
        #         # for each neuron, calculate the probability it spikes in
        #         # each state...
        #         emissions_list += [1/dt * np.stack([
        #             np.mean(X[i:j][state_path==m, :], axis=0) \
        #             for m in range(self.n_components)])]
        #
        # return np.nan_to_num(np.stack(emissions_list))

        # TODO this will depend on the emissions model; how to generalize?

        raise NotImplementedError

    def _compute_shuffle_posterior(self, X, lengths, trial_type):
        # """Doc string"""
        # np.random.seed()
        #
        # shuffle_trials = lambda m: np.dstack(
        #     [m[np.random.permutation(m.shape[0]), :, x] \
        #          for x in range(m.shape[-1])])
        #
        # sX = X.copy().reshape(len(lengths), -1, X.shape[-1])
        # sX = shuffle_trials(sX).reshape(-1, sX.shape[-1])
        #
        # return self.predict_proba(sX, lengths=lengths)

        # TODO this will depend on the emissions model; how to generalize?

        raise NotImplementedError

    def calc_posterior_ci(self, X, lengths, percentile=95, n_shuffle=10,
                          n_processes=1):
        """Calculate null confidence intervals on the posterior probabilities
        of the hidden states. This is done by shuffling the trials for each
        neuron independently to destroy neuron-neuron correlations, while
        leaving their average firing rate statistics intact. Results are stored
        as an attribute of the model, as:
            self.posterior_ci_

        Parameters
        ----------
        X : array of shape (n_samples, n_features)

        lengths : array of shape (n_trials,)
            Note sum(lengths)==n_samples

        percentile : float, optional
            Percentile of the null distribution to store, default 95.

        n_shuffle : int, optional
            Number of shuffle datasets to generate, default 10.

        n_processes : int, optional
            Spawn a parallel pool with multiple workers, default 1.
        """
        # p = Pool(n_processes) if n_processes > 1 else None
        #
        # if p:
        #     shuffle_post = p.map(lambda x: self._compute_shuffle_posterior(
        #         X, lengths), np.arange(n_shuffle))
        # else:
        #     shuffle_post = map(lambda x: self._compute_shuffle_posterior(
        #         X, lengths), np.arange(n_shuffle))
        #
        # all_shuffles = np.concatenate(shuffle_post, axis=0)
        # self.posterior_ci_ = [np.nanpercentile(all_shuffles[:, x], percentile)
        #                       for x in range(all_shuffles.shape[1])]

        # TODO this will depend on the emissions model; how to generalize?
        raise NotImplementedError


class PoissonHMM(BaseMixin, _BaseHMM, ScoringMixin):

    """Inference for Hidden Markov models with independent Poisson emissions.

    Parameters
    ----------
    n_components : int
        Number of states.
    startprob_prior : array, shape (n_components, )
        Initial state occupation prior distribution. Optional.
    transmat_prior : array, shape (n_components, n_components)
        Matrix of prior transition probabilities between states. Optional.
    means_prior : float
        Optional.
    algorithm : string
        Decoder algorithm. Must be one of "viterbi" or "map".
        Defaults to "viterbi".
    random_state : RandomState or an int seed
        A random number generator instance.
    n_iter : int, optional
        Maximum number of iterations to perform.
    tol : float, optional
        Convergence threshold. EM will stop if the gain in log-likelihood
        relative to initial log-likelihood is below this value.
    verbose : bool, optional
        When ``True`` per-iteration convergence reports are printed
        to :data:`sys.stderr`. You can diagnose convergence via the
        :attr:`monitor_` attribute.
    params : string, optional
        Controls which parameters are updated in the training process.  Can
        contain any combination of 's' for startprob, 't' for transmat, and
        'm' for means of the observation Poisson distributions. Defaults to all
        parameters.
    init_params : string, optional
        Controls which parameters are initialized prior to training. See params
        for options and defaults.
    k_means_init : bool, optional
        When ``True`` emission matrix means are initialized by a modified
        k-means clustering of the observations.
    combine_obs_seq :
        NotImplemented

    Attributes
    ----------
    monitor_ : ConvergenceMonitor
        Monitor object used to check the convergence of EM.
    startprob_ : array, shape (n_components, )
        Initial state occupation distribution.
    transmat_ : array, shape (n_components, n_components)
        Matrix of transition probabilities between states.
    means_ : array, shape (n_components, n_features)
        Matrix of means for the observations Poisson distributions.
    """

    def __init__(self, n_components=1, startprob_prior=1.0, transmat_prior=1.0,
                 means_prior=0.0001, algorithm="viterbi", random_state=None,
                 n_iter=100, tol=1e-5, verbose=False, params="stm",
                 init_params="stm", kmeans_init=True):
        _BaseHMM.__init__(self, n_components, startprob_prior=startprob_prior,
                          transmat_prior=transmat_prior, algorithm=algorithm,
                          random_state=random_state, n_iter=n_iter, tol=tol,
                          params=params, verbose=verbose,
                          init_params=init_params)
        self.kmeans_init = kmeans_init
        self.means_prior = means_prior

    def _check(self):
        super()._check()

        self.means_ = np.asarray(self.means_)
        self.n_features = self.means_.shape[1]

    def _init(self, X, lengths=None):
        super()._init(X, lengths=lengths)
        self.random_state = check_random_state(self.random_state)

        _, n_features = X.shape
        if hasattr(self, 'n_features') and self.n_features != n_features:
            raise ValueError('Unexpected number of dimensions, got %s but '
                             'expected %s' % (n_features, self.n_features))

        self.n_features = n_features
        if 'm' in self.init_params or not hasattr(self, "means_"):
            if self.kmeans_init:
                kmeans = KMeans(n_clusters=self.n_components,
                                random_state=self.random_state)
                kmeans.fit(X)
                self.means_ = kmeans.cluster_centers_
                self.means_ += 0.1 * self.random_state.rand(
                    self.n_components, self.n_features)
                self.means_[self.means_ < self.means_prior] = self.means_prior

            else:
                self.means_ = self.random_state.rand(self.n_components,
                                                     self.n_features)

    def _compute_log_likelihood(self, X):
        # we should store the factorial of X so that we don't have to keep
        # recomputing it
        term1 = -1 * np.ones((X.shape[0], self.n_components)) * \
                            np.sum(self.means_, axis=1)
        term2 = np.dot(X, np.log(self.means_.T))
        term3 = np.sum(np.log(factorial(X)), axis=1).reshape(-1, 1)
        return term1 + term2 - term3

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        stats['post'] = np.zeros(self.n_components)
        stats['obs'] = np.zeros((self.n_components, self.n_features))
        return stats

    def _accumulate_sufficient_statistics(self, stats, X, framelogprob,
                                          posteriors, fwdlattice, bwdlattice):
        super()._accumulate_sufficient_statistics(
            stats, X, framelogprob, posteriors, fwdlattice, bwdlattice)

        if 'm' in self.params:
            stats['obs'] += np.dot(posteriors.T, X)
            stats['post'] += np.sum(posteriors, axis=0)

    def _do_mstep(self, stats):
        super()._do_mstep(stats)

        denom = stats['post'][:, np.newaxis]

        if 'm' in self.params:
            self.means_ = stats['obs'] / denom

            # regularize
            self.means_[self.means_ < self.means_prior] = self.means_prior

    def _generate_sample_from_state(self, state, random_state=None):
        raise NotImplementedError
        # random_state = check_random_state(random_state)
        # return [(self.emissionprob_[state, :] > random_state.rand()).argmax()]


class BernoulliHMM(BaseMixin, _BaseHMM, ScoringMixin):

    """Inference for Hidden Markov models with independent Bernoulli emissions.

    Parameters
    ----------
    n_components : int
        Number of states.
    startprob_prior : array, shape (n_components, )
        Initial state occupation prior distribution.
    transmat_prior : array, shape (n_components, n_components)
        Matrix of prior transition probabilities between states.
    algorithm : string
        Decoder algorithm. Must be one of "viterbi" or "map".
        Defaults to "viterbi".
    random_state: RandomState or an int seed
        A random number generator instance.
    n_iter : int, optional
        Maximum number of iterations to perform.
    tol : float, optional
        Convergence threshold. EM will stop if the gain in log-likelihood
        relative to initial log-likelihood is below this value.
    verbose : bool, optional
        When ``True`` per-iteration convergence reports are printed
        to :data:`sys.stderr`. You can diagnose convergence via the
        :attr:`monitor_` attribute.
    params : string, optional
        Controls which parameters are updated in the training
        process.  Can contain any combination of 's' for startprob,
        't' for transmat, and other characters for subclass-specific
        emission parameters. Defaults to all parameters.
    init_params : string, optional
        Controls which parameters are initialized prior to
        training.  Can contain any combination of 's' for
        startprob, 't' for transmat, and other characters for
        subclass-specific emission parameters. Defaults to all
        parameters.
    k_means_init : bool, optional
        When ``True`` emission matrix probabilities are initialized by a
        modified k-means clustering of the observations.

    Attributes
    ----------
    monitor_ : ConvergenceMonitor
        Monitor object used to check the convergence of EM.
    startprob_ : array, shape (n_components, )
        Initial state occupation distribution.
    transmat_ : array, shape (n_components, n_components)
        Matrix of transition probabilities between states.
    emissionprob_ : array, shape (n_components, n_features)
    """

    def __init__(self, n_components=1, startprob_prior=1.0, transmat_prior=1.0,
                 emission_prior=0.0065, algorithm="viterbi", random_state=None,
                 n_iter=100, tol=1e-5, verbose=False, params="ste",
                 init_params="ste", kmeans_init=True,
                 relative_convergence=True):
        _BaseHMM.__init__(self, n_components, startprob_prior=startprob_prior,
                          transmat_prior=transmat_prior, algorithm=algorithm,
                          random_state=random_state, n_iter=n_iter, tol=tol,
                          params=params, verbose=verbose,
                          init_params=init_params)
        self.kmeans_init = kmeans_init
        self.emission_prior = emission_prior
        if relative_convergence:
            self.monitor_ = RelativeConvergenceMonitor(self.tol, self.n_iter,
                                                       self.verbose)

    def _check(self):
        super()._check()

        self.emissionprob_ = np.asarray(self.emissionprob_)
        self.n_features = self.emissionprob_.shape[1]

    def _init(self, X, lengths=None):
        super()._init(X, lengths=lengths)
        self.random_state = check_random_state(self.random_state)

        _, n_features = X.shape
        if hasattr(self, 'n_features') and self.n_features != n_features:
            raise ValueError('Unexpected number of dimensions, got %s but '
                             'expected %s' % (n_features, self.n_features))

        self.n_features = n_features
        if 'e' in self.init_params or not hasattr(self, "emissionprob_"):
            if self.kmeans_init:
                kmeans = KMeans(n_clusters=self.n_components,
                                random_state=self.random_state)
                kmeans.fit(np.nan_to_num(sp.zscore(X)))
                self.emissionprob_ = kmeans.cluster_centers_
                self.emissionprob_[self.emissionprob_ < 0] = 0
                self.emissionprob_ /= (np.max(self.emissionprob_) * 1.25)
                self.emissionprob_ += 0.1 * self.random_state.rand(
                    self.n_components, self.n_features)
                self.emissionprob_ /= (np.max(self.emissionprob_) * 1.25)
            else:
                self.emissionprob_ = self.random_state.rand(self.n_components,
                                                            self.n_features)

    def _compute_log_likelihood(self, X):
        term1 = np.nan_to_num(np.dot(X == 1, np.log(self.emissionprob_.T)))
        term2 = np.nan_to_num(np.dot(X == 0, np.log(1 - self.emissionprob_.T)))
        return term1 + term2

    def _generate_sample_from_state(self, state, random_state=None):
        random_state = check_random_state(random_state)
        return [(self.emissionprob_[state, :] > random_state.rand()).argmax()]

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        stats['obs'] = np.zeros((self.n_components, self.n_features))
        stats['post'] = np.zeros((self.n_components))
        return stats

    def _accumulate_sufficient_statistics(self, stats, X, framelogprob,
                                          posteriors, fwdlattice, bwdlattice):
        super()._accumulate_sufficient_statistics(
            stats, X, framelogprob, posteriors, fwdlattice, bwdlattice)

        if 'e' in self.params:
            stats['obs'] += np.dot(posteriors.T, X)
            stats['post'] += np.sum(posteriors, axis=0)

    def _do_mstep(self, stats):
        super()._do_mstep(stats)

        if 'e' in self.params:
            denom = stats['post'][:, np.newaxis]
            self.emissionprob_ = stats['obs'] / denom

            # regularize
            self.emissionprob_[self.emissionprob_ < self.emission_prior] = \
                self.emission_prior

    def set_params(**params):
        """Set the model parameters a priori"""
        # will be useful for storing model information
        raise NotImplementedError



import numpy as np


def find_intervals(bool_array, circular=False):
    """Get the start and stop (exclusive) index of all intervals of True in an
    array of booleans.

    Parameters
    ----------
    bool_array : array
    circular : bool, optional
        If True, intervals can wrap around the end of the array
    """

    bool_array = np.asarray(bool_array, dtype='int')
    borders = np.concatenate([[0], np.diff(bool_array)])

    starts = np.where(borders == 1)[0]
    ends = np.where(borders == -1)[0]
    max_bin = len(borders)

    intervals = []

    if not starts.size > 0:
        # there are no positive borders.
        # the interval must start at bin 0, and end at the only negative border
        intervals.append([0, ends[0]])

    elif not ends.size > 0:
        # there are no negative borders.
        # the interval must end at bin 0, and start at the only positive border
        intervals.append([starts[0], len(borders)])

    else:
        for s in starts:
            if s > ends[-1]:
                # the interval started after any ends. if circular, the end
                # may wrap around or is max_bin. if not circular, it is max_bin
                if circular & (ends[0] < starts[0]):
                    intervals.append([s, ends[0]])
                else:
                    intervals.append([s, max_bin])
            else:
                # otherwise find the closest end to this start
                valid_ends = ends[ends > s]
                intervals.append([s, valid_ends[np.argmin(valid_ends - s)]])

        if ends[0] < starts[0]:
            # there's an orphaned end (it occurs before any start)
            # if circular, the start may have wrapped around, or it is 0
            if circular & (starts[-1] > (ends[-1])):
                pass
            else:
                intervals.append([0, ends[0]])

    return intervals

def circular_centroid(series):
    n_samples = len(series)
    theta = np.linspace(0, 1, n_samples)*2*np.pi
    xi = np.cos(theta)
    zeta = np.sin(theta)
    
    xi_bar = xi.dot(series) / np.sum(series)
    zeta_bar = zeta.dot(series) / np.sum(series)
    
    theta_bar = np.arctan2(-zeta_bar, -xi_bar) + np.pi
    
    return theta_bar / (2*np.pi)
    
